import os
import platform
import shutil

import pytest

from conan.api.subapi.workspace import WorkspaceAPI
from conan.test.utils.mocks import ConanFileMock
from conan.test.utils.tools import TestClient
from conan.tools.files import replace_in_file

WorkspaceAPI.TEST_ENABLED = "will_break_next"


@pytest.mark.tool("cmake")
def test_build():
    # This is not using the meta-project at all
    c = TestClient()
    c.run("new cmake_lib -d name=mymath")
    c.run("create . -tf=")

    c.save({}, clean_first=True)
    c.run("new workspace -d requires=mymath/0.1")
    c.run("workspace build")
    assert "conanfile.py (app1/0.1): Calling build()" in c.out
    assert "conanfile.py (app1/0.1): Running CMake.build()" in c.out
    # it works without failing


# The workspace CMake needs at least 3.25 for find_package to work
@pytest.mark.tool("cmake", "3.27")
def test_metabuild():
    # This is using the meta-project
    c = TestClient()
    c.run("new cmake_lib -d name=mymath")
    c.run("create . -tf=")

    c.save({}, clean_first=True)
    c.run("new workspace -d requires=mymath/0.1")
    c.run("workspace super-install")
    assert os.path.exists(os.path.join(c.current_folder, "CMakeUserPresets.json"))
    build_folder = "build/Release" if platform.system() != "Windows" else "build"
    assert os.path.exists(os.path.join(c.current_folder, build_folder, "generators"))
    config_preset = "conan-default" if platform.system() == "Windows" else "conan-release"
    c.run_command(f"cmake --preset {config_preset}")
    assert "Conan: Target declared 'mymath::mymath'" in c.out
    assert "Adding project: liba" in c.out
    assert "Adding project: libb" in c.out
    assert "Adding project: app1" in c.out
    c.run_command("cmake --build --preset conan-release")
    # it doesn't fail


# The workspace CMake needs at least 3.25 for find_package to work
@pytest.mark.tool("cmake", "3.27")
def test_new_template_and_different_folder():
    """
    Issue related: https://github.com/conan-io/conan/issues/18813
    """
    c = TestClient()
    c.run("new workspace")
    shutil.move(os.path.join(c.current_folder, "liba"), os.path.join(c.current_folder, "libX"))
    replace_in_file(ConanFileMock(), os.path.join(c.current_folder, "conanws.yml"),
                    "  - path: liba",
                    "  - path: libX")
    c.run("workspace super-install")
    config_preset = "conan-default" if platform.system() == "Windows" else "conan-release"
    c.run_command(f"cmake --preset {config_preset}")  # it does not fail
    assert "Adding project: liba" in c.out
    assert "Adding project: libb" in c.out
    assert "Adding project: app1" in c.out


# The workspace CMake needs at least 3.25 for find_package to work
@pytest.mark.tool("cmake", "3.27")
def test_super_build_different_layouts():
    """
    https://github.com/conan-io/conan/issues/18875
    """
    c = TestClient()
    c.run("new workspace")
    liba_src = os.path.join(c.current_folder, "liba", "mysrc")
    os.makedirs(liba_src)
    shutil.move(os.path.join(c.current_folder, "liba", "CMakeLists.txt"), liba_src)
    shutil.move(os.path.join(c.current_folder, "liba", "include"), liba_src,)
    shutil.move(os.path.join(c.current_folder, "liba", "src"), liba_src)
    replace_in_file(ConanFileMock(), os.path.join(c.current_folder, "liba", "conanfile.py"),
                    "cmake_layout(self)",
                    "cmake_layout(self, src_folder='mysrc')")

    c.run("workspace super-install")

    config_preset = "conan-default" if platform.system() == "Windows" else "conan-release"
    c.run_command(f"cmake --preset {config_preset}")  # it does not fail
    assert "Adding project: liba" in c.out
    assert "Adding project: libb" in c.out
    assert "Adding project: app1" in c.out



import textwrap

import pytest

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


class TestConanfileErrors:

    def test_copy_error(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            from conan import ConanFile

            class HelloConan(ConanFile):
                name = "hello"
                version = "0.1"
                exports = "*"
                def package(self):
                    self.copy2("*.h", dst="include", src=["include","platform"])
            ''')
        files = {"conanfile.py": conanfile, "test.txt": "Hello world"}
        client.save(files)
        client.run("export . --user=lasote --channel=stable")
        client.run("install --requires=hello/0.1@lasote/stable --build='*'", assert_error=True)
        assert "hello/0.1@lasote/stable: Error in package() method, line 9" in client.out
        assert 'self.copy2("*.h", dst="include", src=["include","platform"]' in client.out
        assert "'HelloConan' object has no attribute 'copy2'" in client.out

    def test_copy_error2(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            from conan import ConanFile

            class HelloConan(ConanFile):
                name = "hello"
                version = "0.1"
                exports = "*"
                def package(self):
                    self.copy("*.h", dst="include", src=["include","platform"])
            ''')
        files = {"conanfile.py": conanfile, "test.txt": "Hello world"}
        client.save(files)
        client.run("export . --user=lasote --channel=stable")
        client.run("install --requires=hello/0.1@lasote/stable --build='*'", assert_error=True)
        assert "hello/0.1@lasote/stable: Error in package() method, line 9" in client.out
        assert 'self.copy("*.h", dst="include", src=["include","platform"]' in client.out
        # It results that the error is different in different Python2/3 and OSs
        # assert "'list' object has no attribute 'replace'" in client.out

    def test_package_info_error(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            from conan import ConanFile

            class HelloConan(ConanFile):
                name = "hello"
                version = "0.1"
                exports = "*"
                def package_info(self):
                    self.copy2()
            ''')
        files = {"conanfile.py": conanfile, "test.txt": "Hello world"}
        client.save(files)
        client.run("export . --user=lasote --channel=stable")
        client.run("install --requires=hello/0.1@lasote/stable --build='*'", assert_error=True)
        assert "hello/0.1@lasote/stable: Error in package_info() method, line 9" in client.out
        assert 'self.copy2()' in client.out
        assert "'HelloConan' object has no attribute 'copy2'" in client.out

    def test_config_error(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            from conan import ConanFile

            class HelloConan(ConanFile):
                name = "hello"
                version = "0.1"
                exports = "*"
                def configure(self):
                    self.copy2()
            ''')
        files = {"conanfile.py": conanfile, "test.txt": "Hello world"}
        client.save(files)
        client.run("export . --user=lasote --channel=stable")
        client.run("install --requires=hello/0.1@lasote/stable --build='*'", assert_error=True)

        assert "ERROR: hello/0.1@lasote/stable: Error in configure() method, line 9" in client.out
        assert "self.copy2()" in client.out
        assert "AttributeError: 'HelloConan' object has no attribute 'copy2'""" in client.out

    def test_source_error(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            from conan import ConanFile

            class HelloConan(ConanFile):
                name = "hello"
                version = "0.1"
                exports = "*"
                def source(self):
                    self.copy2()
            ''')
        files = {"conanfile.py": conanfile, "test.txt": "Hello world"}
        client.save(files)
        client.run("export . --user=lasote --channel=stable")
        client.run("install --requires=hello/0.1@lasote/stable --build='*'", assert_error=True)
        assert "hello/0.1@lasote/stable: Error in source() method, line 9" in client.out
        assert 'self.copy2()' in client.out
        assert "'HelloConan' object has no attribute 'copy2'" in client.out

    def test_duplicate_requires(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            [requires]
            foo/0.1@user/testing
            foo/0.2@user/testing
            ''')
        files = {"conanfile.txt": conanfile}
        client.save(files)
        client.run("install . --build='*'", assert_error=True)
        assert "ERROR: Duplicated requirement" in client.out

    def test_duplicate_requires_py(self):
        client = TestClient(light=True)
        conanfile = textwrap.dedent('''
            from conan import ConanFile

            class HelloConan(ConanFile):
                name = "hello"
                version = "0.1"
                requires = "foo/0.1@user/testing", "foo/0.2@user/testing"
            ''')
        files = {"conanfile.py": conanfile}
        client.save(files)
        client.run("export .", assert_error=True)
        assert "Duplicated requirement" in client.out


class TestWrongMethods:
    # https://github.com/conan-io/conan/issues/12961
    @pytest.mark.parametrize("requires", ["requires", "tool_requires",
                                          "test_requires", "build_requires"])
    def test_wrong_method_requires(self, requires):
        """ this is expected to be a relatively frequent user error, and the trace was
        very ugly and debugging complicated
        """
        c = TestClient(light=True)
        conanfile = textwrap.dedent(f"""
            from conan import ConanFile
            class Pkg(ConanFile):
                def {requires}(self):
                    pass
            """)
        c.save({"conanfile.py": conanfile})
        c.run("install .", assert_error=True)
        expected = "requirements" if requires == "requires" else "build_requirements"
        assert f" Wrong '{requires}' definition, did you mean '{expected}()'?" in c.out


def test_notduplicate_requires_py():
    client = TestClient(light=True)
    conanfile = textwrap.dedent('''
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.1"
            requires = "foo/0.1@user/testing"
            build_requires = "foo/0.2@user/testing"
        ''')
    files = {"conanfile.py": conanfile}
    client.save(files)
    client.run("export .")
    assert "hello/0.1: Exported" in client.out


def test_requirements_change_options():
    c = TestClient(light=True)
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.1"

            def requirements(self):
                self.options["mydep"].myoption = 3
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .", assert_error=True)
    assert "ERROR: hello/0.1: Dependencies options were defined incorrectly." in c.out


@pytest.mark.parametrize("property_name", ["libdir", "bindir", "includedir"])
@pytest.mark.parametrize("property_content", [[], ["mydir1", "mydir2"]])
def test_shorthand_bad_interface(property_name, property_content):
    c = TestClient(light=True)
    conanfile = textwrap.dedent(f"""
        from conan import ConanFile

        class HelloConan(ConanFile):
            name = "hello"
            version = "0.1"

            def package_info(self):
                self.cpp_info.{property_name}s = {property_content}
                self.output.info(self.cpp_info.{property_name})
        """)
    c.save({"conanfile.py": conanfile})
    c.run("create .", assert_error=True)
    if property_content:
        assert (f"The {property_name} property is undefined because "
                f"{property_name}s has more than one element.") in c.out
    else:
        assert (f"The {property_name} property is undefined because "
                f"{property_name}s is empty.") in c.out


def test_consumer_unexpected():
    tc = TestClient(light=True)
    cmake_conanfile = textwrap.dedent("""
    from conan import ConanFile

    class CMake(ConanFile):
        name = "cmake"
        version = "1.0"
        package_type = "application"

        def package_info(self):
            self.output.info("cmake/1.0 -> " + str(self._conan_is_consumer))
    """)
    tc.save({
             "cmake/conanfile.py": cmake_conanfile,
             "dep1/conanfile.py": GenConanfile("dep1", "1.0"),
             "conanfile.py": GenConanfile("pkg", "1.0").with_requires("dep1/1.0"),
             "profile": "include(default)\n[tool_requires]\n*: cmake/1.0\n"})
    tc.run("export dep1")
    tc.run("create cmake")
    tc.run("create . -b=missing -pr=profile")
    assert "cmake/1.0 -> True" not in tc.out


@pytest.mark.parametrize("languages", [
    "['C', 'CXX']",
    "['CXX']",
    "'CXX'",
])
def test_language_unexpected(languages):
    tc = TestClient(light=True)
    tc.save({"conanfile.py": GenConanfile().with_class_attribute(f"languages = {languages}")})
    tc.run("inspect .", assert_error=True)
    assert "Only 'C' and 'C++' languages are allowed in 'languages' attribute" in tc.out


def test_empty_languages():
    tc = TestClient(light=True)
    tc.save({"conanfile.py": GenConanfile().with_class_attribute("languages = []")})
    tc.run("inspect .")
    assert "Only 'C' and 'C++' languages are allowed in 'languages' attribute" not in tc.out



import platform


def check_binary(t):
    # FIXME: Some values are hardcoded to match the CI setup
    host_arch = t.get_default_host_profile().settings['arch']
    arch_macro = {
        "gcc": {"armv8": "__aarch64__", "x86_64": "__x86_64__"},
        "msvc": {"armv8": "_M_ARM64", "x86_64": "_M_X64"}
    }
    if platform.system() == "Darwin":
        assert f"main {arch_macro['gcc'][host_arch]} defined" in t.out
        assert "main __apple_build_version__" in t.out
        assert "main __clang_major__17" in t.out
        # TODO: check why __clang_minor__ seems to be not defined in XCode 12
        # commented while migrating to XCode12 CI
        # assert ("main __clang_minor__0" in t.out
    elif platform.system() == "Windows":
        assert f"main {arch_macro['msvc'][host_arch]} defined" in t.out
        assert "main _MSC_VER19" in t.out
        assert "main _MSVC_LANG2014" in t.out
    elif platform.system() == "Linux":
        assert f"main {arch_macro['gcc'][host_arch]} defined" in t.out
        assert "main __GNUC__9" in t.out



import json
import textwrap

import pytest

from conan.test.utils.env import environment_update
from conan.test.utils.tools import TestClient, GenConanfile, TestServer


class TestCompatibleIDsTest:

    def test_compatible_setting_no_binary(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile

           class Pkg(ConanFile):
               settings = "os", "compiler"
               def compatibility(self):
                    if self.settings.compiler == "gcc" and self.settings.compiler.version == "4.9":
                        return [{"settings": [("compiler.version", v)]}
                                for v in ("4.8", "4.7", "4.6")]
               def package_info(self):
                   self.output.info("PackageInfo!: Gcc version: %s!"
                                    % self.settings.compiler.version)
           """)
        profile = textwrap.dedent("""
           [settings]
           os = Linux
           compiler=gcc
           compiler.version=4.9
           compiler.libcxx=libstdc++
           """)
        client.save({"conanfile.py": conanfile,
                     "myprofile": profile})
        # Create package with gcc 4.8
        client.run("export . --name=pkg --version=0.1 --user=user --channel=stable")
        assert ("pkg/0.1@user/stable: Exported: "
                "pkg/0.1@user/stable#d165eb4bcdd1c894a97d2a212956f5fe") in client.out
        client.run("export . --name=lib --version=0.1 --user=user --channel=stable")

        # package can be used with a profile gcc 4.9 falling back to 4.8 binary
        client.save({"conanfile.py": GenConanfile().with_requires("pkg/0.1@user/stable", "lib/0.1@user/stable")})
        # No fallback
        client.run("install . -pr=myprofile --build=missing -u=lib")
        assert "pkg/0.1@user/stable: PackageInfo!: Gcc version: 4.9!" in client.out
        client.assert_listed_binary({"pkg/0.1@user/stable":
                                     ("1ded27c9546219fbd04d4440e05b2298f8230047", "Build")})
        assert "lib/0.1@user/stable: Compatible configurations not found in cache, checking servers" not in client.out
        assert "pkg/0.1@user/stable: Compatible configurations not found in cache, checking servers" in client.out

    def test_compatible_setting_no_user_channel(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                settings = "os", "compiler"
                def compatibility(self):
                    if self.settings.compiler == "gcc" and self.settings.compiler.version == "4.9":
                        return [{"settings": [("compiler.version", v)]}
                                for v in ("4.8", "4.7", "4.6")]
            """)
        profile = textwrap.dedent("""
            [settings]
            os = Linux
            compiler=gcc
            compiler.version=4.9
            compiler.libcxx=libstdc++
            """)
        client.save({"conanfile.py": conanfile,
                     "myprofile": profile})

        # No user/channel
        client.run("create . --name=pkg --version=0.1 -pr=myprofile -s compiler.version=4.8")
        package_id = client.created_package_id("pkg/0.1")

        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1")})
        client.run("install . -pr=myprofile")
        client.assert_listed_binary({"pkg/0.1": (package_id, "Cache")})
        assert "pkg/0.1: Already installed!" in client.out

    def test_compatible_option(self):
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                options = {"optimized": [1, 2, 3]}
                default_options = {"optimized": 1}

                def compatibility(self):
                    return [{"options": [("optimized", v)]}
                            for v in range(int(self.options.optimized), 0, -1)]

                def package_info(self):
                    self.output.info("PackageInfo!: Option optimized %s!"
                                     % self.options.optimized)
            """)
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=stable")
        package_id = client.created_package_id("pkg/0.1@user/stable")
        assert f"pkg/0.1@user/stable: Package '{package_id}' created" in client.out

        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1@user/stable")})
        client.run("install . -o pkg/*:optimized=2 -vv")
        # Information messages
        missing_id = "0a8157f8083f5ece34828d27fb2bf5373ba26366"
        assert "pkg/0.1@user/stable: PackageInfo!: Option optimized 1!" in client.out
        assert (f"pkg/0.1@user/stable: Compatible package ID {missing_id} "
                f"equal to the default package ID") in client.out
        assert f"pkg/0.1@user/stable: Main binary package '{missing_id}' missing" in client.out
        assert f"Found compatible package '{package_id}'" in client.out
        # checking the resulting dependencies
        client.assert_listed_binary({"pkg/0.1@user/stable": (package_id, "Cache")})
        assert "pkg/0.1@user/stable: Already installed!" in client.out
        client.run("install . -o pkg/*:optimized=3")
        client.assert_listed_binary({"pkg/0.1@user/stable": (package_id, "Cache")})
        assert "pkg/0.1@user/stable: Already installed!" in client.out

    def test_package_id_consumers(self):
        # If we fallback to a different binary upstream and we are using a "package_revision_mode"
        # the current package should have a different binary package ID too.
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                settings = "os", "compiler"
                def compatibility(self):
                    return [{"settings": [("compiler.version", "4.8")]}]
                def package_info(self):
                    self.output.info("PackageInfo!: Gcc version: %s!"
                                     % self.settings.compiler.version)
            """)
        profile = textwrap.dedent("""
            [settings]
            os = Linux
            compiler=gcc
            compiler.version=4.9
            compiler.libcxx=libstdc++
            """)
        client.save_home({"global.conf":
                          "core.package_id:default_unknown_mode=recipe_revision_mode"})
        client.save({"conanfile.py": conanfile,
                     "myprofile": profile})
        # Create package with gcc 4.8
        client.run("create . --name=pkg --version=0.1 --user=user --channel=stable "
                   "-pr=myprofile -s compiler.version=4.8")
        package_id = client.created_package_id("pkg/0.1@user/stable")
        assert f"pkg/0.1@user/stable: Package '{package_id}' created" in client.out

        # package can be used with a profile gcc 4.9 falling back to 4.8 binary
        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1@user/stable")})
        client.run("create . --name=consumer --version=0.1 --user=user --channel=stable -pr=myprofile")
        assert "pkg/0.1@user/stable: PackageInfo!: Gcc version: 4.8!" in client.out
        client.assert_listed_binary({"pkg/0.1@user/stable": (package_id, "Cache")})
        assert "pkg/0.1@user/stable: Already installed!" in client.out
        consumer_id = "96465a24a53766aaac28e270d196db295e2fd22a"
        client.assert_listed_binary({"consumer/0.1@user/stable": (consumer_id, "Build")})
        assert f"consumer/0.1@user/stable: Package '{consumer_id}' created" in client.out

        # Create package with gcc 4.9
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=stable -pr=myprofile")
        package_id = "1ded27c9546219fbd04d4440e05b2298f8230047"
        assert f"pkg/0.1@user/stable: Package '{package_id}' created" in client.out

        # Consume it
        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1@user/stable")})
        client.run("create . --name=consumer --version=0.1 --user=user --channel=stable -pr=myprofile")
        assert "pkg/0.1@user/stable: PackageInfo!: Gcc version: 4.9!" in client.out
        client.assert_listed_binary({"pkg/0.1@user/stable": (f"{package_id}", "Cache")})
        assert "pkg/0.1@user/stable: Already installed!" in client.out
        consumer_id = "41bc915fa380e9a046aacbc21256fcb46ad3179d"
        client.assert_listed_binary({"consumer/0.1@user/stable": (consumer_id, "Build")})
        assert f"consumer/0.1@user/stable: Package '{consumer_id}' created" in client.out

    def test_build_missing(self):
        # https://github.com/conan-io/conan/issues/6133
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Conan(ConanFile):
                settings = "os"

                def compatibility(self):
                    if self.settings.os == "Windows":
                        return [{"settings": [("os", "Linux")]}]
                """)

        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -s os=Linux")
        package_id = client.created_package_id("pkg/0.1@user/testing")
        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1@user/testing")})
        client.run("install . -s os=Windows --build=missing")
        client.assert_listed_binary({"pkg/0.1@user/testing": (package_id, "Cache")})
        assert "pkg/0.1@user/testing: Already installed!" in client.out

    def test_compatible_package_python_requires(self):
        # https://github.com/conan-io/conan/issues/6609
        client = TestClient()
        client.save({"conanfile.py": GenConanfile()})
        client.run("export . --name=tool --version=0.1")
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Conan(ConanFile):
                settings = "os"
                python_requires = "tool/0.1"

                def compatibility(self):
                    if self.settings.os == "Windows":
                        return [{"settings": [("os", "Linux")]}]
                """)

        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing -s os=Linux")
        package_id = client.created_package_id("pkg/0.1@user/testing")
        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1@user/testing")})
        client.run("install . -s os=Windows")
        client.assert_listed_binary({"pkg/0.1@user/testing": (package_id, "Cache")})
        assert "pkg/0.1@user/testing: Already installed!" in client.out

    def test_compatible_lockfile(self):
        # https://github.com/conan-io/conan/issues/9002
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                settings = "os"
                def compatibility(self):
                    if self.settings.os == "Windows":
                        return [{"settings": [("os", "Linux")]}]
                def package_info(self):
                    self.output.info("PackageInfo!: OS: %s!" % self.settings.os)
            """)

        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 -s os=Linux")
        assert "pkg/0.1: PackageInfo!: OS: Linux!" in client.out
        assert "pkg/0.1: Package '9a4eb3c8701508aa9458b1a73d0633783ecc2270' built" in client.out

        client.save({"conanfile.py": GenConanfile().with_require("pkg/0.1")})
        client.run("lock create . -s os=Windows --lockfile-out=deps.lock")
        client.run("install . -s os=Windows --lockfile=deps.lock")
        assert "pkg/0.1: PackageInfo!: OS: Linux!" in client.out
        assert "9a4eb3c8701508aa9458b1a73d0633783ecc2270" in client.out
        assert "pkg/0.1: Already installed!" in client.out

    def test_compatible_diamond(self):
        # https://github.com/conan-io/conan/issues/9880
        client = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile
            class Pkg(ConanFile):
                {}
                settings = "build_type"
                def compatibility(self):
                    if self.settings.build_type == "Debug":
                       return [{{"settings": [("build_type", "Release")]}}]
                """)

        private = """def requirements(self):
        self.requires("pkga/0.1", visible=False)
        """
        client.save({"pkga/conanfile.py": conanfile.format(""),
                     "pkgb/conanfile.py": conanfile.format(private),
                     "pkgc/conanfile.py": conanfile.format('requires = "pkga/0.1"'),
                     "pkgd/conanfile.py": conanfile.format('requires = "pkgb/0.1", "pkgc/0.1"')
                     })
        client.run("create pkga --name=pkga --version=0.1 -s build_type=Release")
        client.run("create pkgb --name=pkgb --version=0.1 -s build_type=Release")
        client.run("create pkgc --name=pkgc --version=0.1 -s build_type=Release")

        client.run("install pkgd -s build_type=Debug")
        client.assert_listed_binary({"pkga/0.1":
                                    ("efa83b160a55b033c4ea706ddb980cd708e3ba1b", "Cache")})


class TestNewCompatibility:

    def test_compatible_setting(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class Pkg(ConanFile):
                name = "pkg"
                version = "0.1"
                settings = "os", "compiler"

                def compatibility(self):
                    if self.settings.compiler == "gcc" and self.settings.compiler.version == "4.9":
                        return [{"settings": [("compiler.version", v)]}
                                for v in ("4.8", "4.7", "4.6")]

                def package_info(self):
                    self.output.info("PackageInfo!: Gcc version: %s!"
                                     % self.settings.compiler.version)
            """)
        profile = textwrap.dedent("""
            [settings]
            os = Linux
            compiler=gcc
            compiler.version=4.9
            compiler.libcxx=libstdc++
            """)
        c.save({"conanfile.py": conanfile,
                "myprofile": profile})
        # Create package with gcc 4.8
        c.run("create .  -pr=myprofile -s compiler.version=4.8")
        package_id = "c0c95d81351786c6c1103566a27fb1c1f78629ac"
        assert f"pkg/0.1: Package '{package_id}' created" in c.out

        # package can be used with a profile gcc 4.9 falling back to 4.8 binary
        c.save({"conanfile.py": GenConanfile().with_require("pkg/0.1")})
        c.run("install . -pr=myprofile")
        assert "pkg/0.1: PackageInfo!: Gcc version: 4.8!" in c.out
        c.assert_listed_binary({"pkg/0.1": (f"{package_id}", "Cache")})
        assert "pkg/0.1: Already installed!" in c.out

    def test_compatibility_remove_package_id(self):
        # https://github.com/conan-io/conan/issues/13727
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class PdfiumConan(ConanFile):
                name = "pdfium"
                version = "2020.9"
                settings = "os", "compiler", "arch", "build_type"
                build_policy = "never"

                def compatibility(self):
                    result = []
                    if self.info.settings.build_type == "Debug":
                        result.append({"settings": [("build_type", "Release")]})
                    return result

                def package_id(self):
                    del self.info.settings.compiler.runtime
                    del self.info.settings.compiler.runtime_type
            """)
        profile = textwrap.dedent("""
            [settings]
            os = Windows
            compiler=msvc
            compiler.version=192
            compiler.runtime=dynamic
            build_type=Release
            arch=x86_64
            """)
        c.save({"conanfile.py": conanfile,
                "myprofile": profile})
        c.run("create .  -pr=myprofile", assert_error=True)
        assert "ERROR: This package cannot be created, 'build_policy=never', " \
               "it can only be 'export-pkg'" in c.out
        c.run("export-pkg . -pr=myprofile")
        c.run("list pdfium/2020.9:*")

        c.run("install --requires=pdfium/2020.9 -pr=myprofile -s build_type=Debug")
        assert "Found compatible package" in c.out

    def test_compatibility_erase_package_id(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
            from conan import ConanFile

            class PdfiumConan(ConanFile):
                name = "diligent-core"
                version = "1.0"
                settings = "compiler"
                options = {"foo": ["no"]}

                def package_id(self):
                    self.info.settings.compiler.runtime = "foobar"
                    self.info.options.foo = "yes"
            """)
        profile = textwrap.dedent("""
            [settings]
            os = Windows
            compiler=msvc
            compiler.version=192
            compiler.runtime=dynamic
            build_type=Release
            arch=x86_64
            """)
        c.save({"conanfile.py": conanfile,
                "myprofile": profile})

        c.run("create . -pr:a=myprofile -s compiler.cppstd=20")
        c.run("install --requires=diligent-core/1.0 -pr:a=myprofile -s compiler.cppstd=17")
        assert "ERROR: Invalid setting 'foobar' is not a valid 'settings.compiler.runtime' value." not in c.out

    def test_compatibility_msvc_and_cppstd(self):
        """msvc 194 would not find compatible packages built with same version but different cppstd
        due to an issue in the msvc fallback compatibility rule."""
        tc = TestClient()
        profile = textwrap.dedent("""
                   [settings]
                   compiler=msvc
                   compiler.version=194
                   compiler.runtime=dynamic
                   """)
        tc.save({"dep/conanfile.py": GenConanfile("dep", "1.0").with_setting("compiler"),
                 "profile": profile})

        tc.run("create dep -pr=profile -s compiler.cppstd=20")
        tc.run("install --requires=dep/1.0 -pr=profile -s compiler.cppstd=17")
        tc.assert_listed_binary({"dep/1.0": ("b6d26a6bc439b25b434113982791edf9cab4d004", "Cache")})

        tc.run("remove * -c")
        tc.run("create dep -pr=profile -s compiler.version=193 -s compiler.cppstd=20")
        tc.run("install --requires=dep/1.0 -pr=profile -s compiler.cppstd=17")
        assert "compiler.cppstd=20, compiler.version=193" in tc.out
        tc.assert_listed_binary({"dep/1.0": ("535899bb58c3ca7d80a380313d31f4729e735d1c", "Cache")})


class TestCompatibleBuild:
    def test_build_compatible(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.build import check_min_cppstd

           class Pkg(ConanFile):
               name = "pkg"
               version = "0.1"
               settings = "os", "compiler"

               def validate(self):
                   check_min_cppstd(self, 14)
            """)
        c.save({"conanfile.py": conanfile})
        settings = "-s os=Windows -s compiler=gcc -s compiler.version=11 " \
                   "-s compiler.libcxx=libstdc++11 -s compiler.cppstd=11"
        c.run(f"create . {settings}", assert_error=True)
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Invalid: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=compatible:&")
        # the one for cppstd=14 is built!!
        c.assert_listed_binary({"pkg/0.1": ("389803bed06200476fcee1af2023d4e9bfa24ff9", "Build")})
        c.run("list *:*")
        assert "compiler.cppstd: 14" in c.out

    def test_build_compatible_cant_build(self):
        # requires c++17 to build, can be consumed with c++14
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.build import check_min_cppstd

           class Pkg(ConanFile):
               name = "pkg"
               version = "0.1"
               settings = "os", "compiler"

               def validate(self):
                   check_min_cppstd(self, 14)

               def validate_build(self):
                   check_min_cppstd(self, 17)
            """)
        c.save({"conanfile.py": conanfile})
        settings = "-s os=Windows -s compiler=gcc -s compiler.version=11 " \
                   "-s compiler.libcxx=libstdc++11  -s compiler.cppstd=11"
        c.run(f"create . {settings}", assert_error=True)
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Invalid: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=missing", assert_error=True)
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Invalid: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=compatible:&")
        # the one for cppstd=17 is built!!
        c.assert_listed_binary({"pkg/0.1": ("58fb8ac6c2dc3e3f837253ce1a6ea59011525866", "Build")})
        c.run("list *:*")
        assert "compiler.cppstd: 17" in c.out

    def test_build_compatible_cant_build2(self):
        # requires c++17 to build, can be consumed with c++11
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.build import check_min_cppstd

           class Pkg(ConanFile):
               name = "pkg"
               version = "0.1"
               settings = "os", "compiler"

               def validate(self):
                   check_min_cppstd(self, 11)

               def validate_build(self):
                   check_min_cppstd(self, 17)
            """)
        c.save({"conanfile.py": conanfile})
        settings = "-s os=Windows -s compiler=gcc -s compiler.version=11 " \
                   "-s compiler.libcxx=libstdc++11  -s compiler.cppstd=11"
        c.run(f"create . {settings}", assert_error=True)
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Cannot build for this configuration: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=missing", assert_error=True)
        # the one for cppstd=17 is built!!
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Cannot build for this configuration: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=compatible:&")
        # the one for cppstd=17 is built!!
        c.assert_listed_binary({"pkg/0.1": ("58fb8ac6c2dc3e3f837253ce1a6ea59011525866", "Build")})
        c.run("list *:*")
        assert "compiler.cppstd: 17" in c.out

    def test_build_compatible_cant_build_only(self):
        # requires c++17 to build, but don't specify consumption
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.build import check_min_cppstd

           class Pkg(ConanFile):
               name = "pkg"
               version = "0.1"
               settings = "os", "compiler"

               def validate_build(self):
                   check_min_cppstd(self, 17)
            """)
        c.save({"conanfile.py": conanfile})
        settings = "-s os=Windows -s compiler=gcc -s compiler.version=11 " \
                   "-s compiler.libcxx=libstdc++11  -s compiler.cppstd=11"
        c.run(f"create . {settings}", assert_error=True)
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Cannot build for this configuration: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=missing", assert_error=True)
        # the one for cppstd=17 is built!!
        c.assert_listed_binary({"pkg/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid")})
        assert "pkg/0.1: Cannot build for this configuration: Current cppstd (11)" in c.out

        c.run(f"create . {settings} --build=compatible:&")
        # the one for cppstd=17 is built!!
        c.assert_listed_binary({"pkg/0.1": ("58fb8ac6c2dc3e3f837253ce1a6ea59011525866", "Build")})
        c.run("list *:*")
        assert "compiler.cppstd: 17" in c.out

    def test_multi_level_build_compatible(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.build import check_min_cppstd

           class Pkg(ConanFile):
               name = "{name}"
               version = "0.1"
               settings = "os", "compiler"
               {requires}

               def validate(self):
                   check_min_cppstd(self, {cppstd})
            """)
        c.save({"liba/conanfile.py": conanfile.format(name="liba", cppstd=14, requires=""),
                "libb/conanfile.py": conanfile.format(name="libb", cppstd=17,
                                                      requires='requires="liba/0.1"')})
        c.run("export liba")
        c.run("export libb")
        settings = "-s os=Windows -s compiler=gcc -s compiler.version=11 " \
                   "-s compiler.libcxx=libstdc++11 -s compiler.cppstd=11"
        c.run(f"install --requires=libb/0.1 {settings}", assert_error=True)
        c.assert_listed_binary({"liba/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Invalid"),
                                "libb/0.1": ("144910d65b27bcbf7d544201f5578555bbd0376e", "Invalid")})
        assert "liba/0.1: Invalid: Current cppstd (11)" in c.out
        assert "libb/0.1: Invalid: Current cppstd (11)" in c.out

        c.run(f"install --requires=libb/0.1 {settings} --build=compatible")
        # the one for cppstd=14 is built!!
        c.assert_listed_binary({"liba/0.1": ("389803bed06200476fcee1af2023d4e9bfa24ff9", "Build"),
                                "libb/0.1": ("8f29f49be3ba2b6cbc9fa1e05432ce928b96ae5d", "Build")})
        c.run("list liba:*")
        assert "compiler.cppstd: 14" in c.out
        c.run("list libb:*")
        assert "compiler.cppstd: 17" in c.out

    def test_multi_level_build_compatible_build_order(self):
        c = TestClient()
        conanfile = textwrap.dedent("""
           from conan import ConanFile
           from conan.tools.build import check_min_cppstd

           class Pkg(ConanFile):
               name = "{name}"
               version = "0.1"
               settings = "os", "compiler"
               {requires}

               def validate_build(self):
                   check_min_cppstd(self, {cppstd})
            """)
        c.save({"liba/conanfile.py": conanfile.format(name="liba", cppstd=14, requires=""),
                "libb/conanfile.py": conanfile.format(name="libb", cppstd=17,
                                                      requires='requires="liba/0.1"')})
        c.run("export liba")
        c.run("export libb")
        settings = "-s os=Windows -s compiler=gcc -s compiler.version=11 " \
                   "-s compiler.libcxx=libstdc++11 -s compiler.cppstd=11"
        c.run(f"graph build-order --requires=libb/0.1 {settings} --format=json", assert_error=True,
              redirect_stdout="build_order.json")
        c.assert_listed_binary({"liba/0.1": ("bb33db23c961978d08dc0cdd6bc786b45b3e5943", "Missing"),
                                "libb/0.1": ("144910d65b27bcbf7d544201f5578555bbd0376e", "Missing")})

        c.run(f"graph build-order --requires=libb/0.1 {settings} --build=compatible "
              "--order-by=configuration --format=json", redirect_stdout="build_order.json")
        bo = json.loads(c.load("build_order.json"))
        liba = bo["order"][0][0]
        assert liba["ref"] == "liba/0.1#c1459d256a9c2d3c49d149fd7c43310c"
        assert liba["info"]["compatibility_delta"] == {"settings": [["compiler.cppstd", "14"]]}
        assert liba["build_args"] == "--requires=liba/0.1 --build=compatible:liba/0.1"
        # Lets make sure the build works too
        c.run(f"install {settings} {liba['build_args']}")
        libb = bo["order"][1][0]
        assert libb["ref"] == "libb/0.1#62bb167aaa5306d1ac757bb817797f9e"
        assert libb["info"]["compatibility_delta"] == {"settings": [["compiler.cppstd", "17"]]}
        assert libb["build_args"] == "--requires=libb/0.1 --build=compatible:libb/0.1"
        # Lets make sure the build works too
        c.run(f"install {settings} {libb['build_args']}")

        # Now lets make sure that build-order-merge works too
        c.run(f"graph build-order --requires=libb/0.1 {settings} -s compiler.version=12 "
              "--build=compatible --order-by=configuration --format=json",
              redirect_stdout="build_order2.json")

        c.run(f"graph build-order-merge --file=build_order.json --file=build_order2.json -f=json",
              redirect_stdout="build_order_merged.json")
        bo = json.loads(c.load("build_order_merged.json"))
        for pkg_index in (0, 1):
            liba = bo["order"][0][pkg_index]
            assert liba["ref"] == "liba/0.1#c1459d256a9c2d3c49d149fd7c43310c"
            assert liba["info"]["compatibility_delta"] == {"settings": [["compiler.cppstd", "14"]]}
            assert liba["build_args"] == "--requires=liba/0.1 --build=compatible:liba/0.1"

        # By recipe also works
        c.run("remove *:* -c")
        c.run(f"graph build-order --requires=libb/0.1 {settings} --build=compatible "
              "--order-by=recipe --format=json", redirect_stdout="build_order.json")
        bo = json.loads(c.load("build_order.json"))
        liba = bo["order"][0][0]
        assert liba["ref"] == "liba/0.1#c1459d256a9c2d3c49d149fd7c43310c"
        pkga = liba["packages"][0][0]
        assert pkga["info"]["compatibility_delta"] == {"settings": [["compiler.cppstd", "14"]]}
        assert pkga["build_args"] == "--requires=liba/0.1 --build=compatible:liba/0.1"
        libb = bo["order"][1][0]
        assert libb["ref"] == "libb/0.1#62bb167aaa5306d1ac757bb817797f9e"
        pkgb = libb["packages"][0][0]
        assert pkgb["info"]["compatibility_delta"] == {"settings": [["compiler.cppstd", "17"]]}
        assert pkgb["build_args"] == "--requires=libb/0.1 --build=compatible:libb/0.1"

        # Now lets make sure that build-order-merge works too
        c.run(f"graph build-order --requires=libb/0.1 {settings} -s compiler.version=12 "
              "--order-by=recipe --build=compatible --format=json",
              redirect_stdout="build_order2.json")
        c.run(f"graph build-order-merge --file=build_order.json --file=build_order2.json -f=json",
              redirect_stdout="build_order_merged.json")
        bo = json.loads(c.load("build_order_merged.json"))
        liba = bo["order"][0][0]
        assert liba["ref"] == "liba/0.1#c1459d256a9c2d3c49d149fd7c43310c"
        for pkg_index in (0, 1):
            pkga = liba["packages"][0][pkg_index]
            assert pkga["info"]["compatibility_delta"] == {"settings": [["compiler.cppstd", "14"]]}
            assert pkga["build_args"] == "--requires=liba/0.1 --build=compatible:liba/0.1"


def test_compatibility_new_setting_forwards_compat():
    """ This test tries to reflect the following scenario:
    - User adds a new setting (libc.version in this case)
    - This setting is forward compatible
    How is it solved with compatibility.py? Like this:
    """
    tc = TestClient()
    tc.save_home({"settings_user.yml": "libc_version: [1, 2, 3]"})
    tc.save({"conanfile.py": GenConanfile("dep", "1.0").with_settings("libc_version", "compiler")})
    # The extra cppstd and compiler versions are for later demonstrations of combinations of settings
    # The cppstd=17 and compiler.version=193 are used thought until the last 2 install calls
    tc.run("create . -s=libc_version=2 -s=compiler.cppstd=17")
    dep_package_id = tc.created_package_id("dep/1.0")
    tc.run("install --requires=dep/1.0 -s=libc_version=3 -s=compiler.cppstd=17", assert_error=True)
    # We can't compile, because the dep is not compatible
    assert "Missing prebuilt package for 'dep/1.0'" in tc.out

    # Let's create a compatibility extensions
    libc_compat = textwrap.dedent("""
        from conan.tools.scm import Version

        def libc_compat(conanfile):
            # Do we have the setting?
            libc_version = conanfile.settings.get_safe("libc_version")
            if libc_version is None:
                return []
            available_libc_versions = conanfile.settings.libc_version.possible_values()
            ret = []
            for possible_libc_version in available_libc_versions:
                if Version(possible_libc_version) < Version(libc_version):
                    ret.append({"libc_version": possible_libc_version})
            return ret
        """)
    compat = tc.load_home("extensions/plugins/compatibility/compatibility.py")
    compat = "from libc_compat import libc_compat\n" + compat
    compat = compat.replace("# Append more factors for your custom compatibility rules here",
                            "factors.append(libc_compat(conanfile))")
    tc.save_home({"extensions/plugins/compatibility/libc_compat.py": libc_compat,
                  "extensions/plugins/compatibility/compatibility.py": compat})

    # Now we try again, this time app will find the compatible dep with libc_version 2
    tc.run("install --requires=dep/1.0 -s=libc_version=3 -s=compiler.cppstd=17")
    assert f"dep/1.0: Found compatible package '{dep_package_id}'" in tc.out

    # And now we try to create the app with libc_version 1, which is still not compatible
    tc.run("install --requires=dep/1.0 -s=libc_version=1 -s=compiler.cppstd=17", assert_error=True)
    assert "Missing prebuilt package for 'dep/1.0'" in tc.out

    # Now we try again, this time app will find the compatible dep with libc_version 2
    # And see how we're also compatible over a different cppstd
    tc.run("install --requires=dep/1.0 -s=libc_version=3 -s=compiler.cppstd=14")
    assert f"dep/1.0: Found compatible package '{dep_package_id}': compiler.cppstd=17, " \
           f"libc_version=2" in tc.out


class TestListOnlyCompatibilityOptimization:

    @pytest.fixture()
    def client(self):
        tc = TestClient(default_server_user=True, light=True)
        compiler_settings = textwrap.dedent("""
                compiler:
                    foo:
                        version: [1]
                        cppstd: [7, 11, 14, 17, 20, 23]""")
        tc.run("version")
        tc.save_home({"settings_user.yml": compiler_settings})
        compat = tc.load_home("extensions/plugins/compatibility/compatibility.py")
        compat = compat.replace("cppstd_possible_values = supported_cppstd(conanfile)",
                                "cppstd_possible_values = [7, 11, 14, 17, 20, 23]")
        tc.save_home({"extensions/plugins/compatibility/compatibility.py": compat})
        return tc

    @pytest.mark.parametrize("update", [True, False])
    def test_remote_compatible_package(self, client, update):
        tc = client
        update_arg = "-u" if update else ""
        compiler_args = "-s compiler=foo -s compiler.version=1"
        tc.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("compiler")})
        tc.run(f"create . {compiler_args} -s=compiler.cppstd=14")
        std14_id = tc.created_layout().reference.package_id
        tc.run(f"create . {compiler_args} -s=compiler.cppstd=17")
        std17_id = tc.created_layout().reference.package_id
        tc.run(f"create . {compiler_args} -s=compiler.cppstd=20")
        std20_id = tc.created_layout().reference.package_id

        tc.run(f"upload pkg/0.1:{std20_id} -r=default -c")
        tc.run(f"upload pkg/0.1:{std14_id} -r=default -c")
        tc.run(f"upload pkg/0.1:{std17_id} -r=default -c")

        tc.run("remove * -c")
        tc.run(f"install --requires=pkg/0.1 {compiler_args} -s=compiler.cppstd=11 {update_arg} "
               "-cc core.graph:compatibility_mode=optimized")
        assert f"Found compatible package '{std14_id}'" in tc.out

        tc.run("remove * -c")
        tc.run(f"remove pkg/0.1:{std17_id} -r=default -c")
        tc.run(f"install --requires=pkg/0.1 {compiler_args} -s=compiler.cppstd=17 {update_arg} "
               "-cc core.graph:compatibility_mode=optimized")
        if not update:
            assert "Found 2 compatible configurations in remotes" in tc.out
        assert f"Found compatible package '{std14_id}'" in tc.out

        tc.run(f"create . {compiler_args} -s=compiler.cppstd=11")
        std11_layout = tc.created_layout()
        std11_id = tc.created_layout().reference.package_id
        tc.run(f"upload pkg/0.1:{std11_id} -r=default -c")
        tc.run("remove * -c")
        tc.run(f"install --requires=pkg/0.1 {compiler_args} -s=compiler.cppstd=17 -vvv {update_arg} "
               "-cc core.graph:compatibility_mode=optimized")
        assert f"Found compatible package '{std11_id}'" in tc.out
        # An HTTP request is made to the server to search for compatible packages
        if not update:
            assert "Found 3 compatible configurations in remotes" in tc.out
            assert f"{std11_layout.reference.ref.revision}/search?list_only=True" in tc.out

    def test_remote_compatible_package_update_cache(self, client):
        tc = client
        compiler_args = "-s compiler=foo -s compiler.version=1"
        tc.save({"conanfile.py": GenConanfile("pkg", "0.1")
                    .with_settings("compiler")
                    .with_import("import os, time")
                    .with_import("from conan.tools.files import save")
                    .with_package("save(self, os.path.join(self.package_folder, 'data.txt'), str(time.time()))")
        })
        tc.run(f"create . {compiler_args} -s=compiler.cppstd=17")
        std17_old_ref = tc.created_layout().reference
        std17_id = std17_old_ref.package_id
        tc.run(f"create . {compiler_args} -s=compiler.cppstd=17")
        std17_new_ref = tc.created_layout().reference
        assert std17_old_ref.revision != std17_new_ref.revision

        tc.run(f"upload pkg/0.1:{std17_id}#latest -r=default -c")
        tc.run(f"remove pkg/0.1:{std17_id}#latest -c")

        tc.run(f"install --requires=pkg/0.1 {compiler_args} -s=compiler.cppstd=11 -r=default -u "
               "-cc core.graph:compatibility_mode=optimized")
        assert "Current package revision is older than the remote one "
        assert f"Found compatible package '{std17_id}'" in tc.out
        assert std17_old_ref.revision not in tc.out
        assert std17_new_ref.revision in tc.out

        tc.run("remove * -c")
        tc.run("remove * -r=default -c")

        tc.run(f"create . {compiler_args} -s=compiler.cppstd=20")
        std20_old_ref = tc.created_layout().reference
        std20_id = std20_old_ref.package_id
        tc.run(f"create . {compiler_args} -s=compiler.cppstd=20")
        std20_new_ref = tc.created_layout().reference
        assert std20_old_ref.revision != std20_new_ref.revision

        tc.run(f"upload pkg/0.1:{std20_id}#* -r=default -c")
        tc.run(f"remove pkg/0.1:{std20_id} -c")

        tc.run(f"install --requires=pkg/0.1 {compiler_args} -s=compiler.cppstd=11 -r=default -u "
               "-cc core.graph:compatibility_mode=optimized")
        assert f"Found compatible package '{std20_id}'" in tc.out
        assert std20_old_ref.revision not in tc.out
        assert std20_new_ref.revision in tc.out

    @pytest.mark.parametrize("enable", [True, False])
    @pytest.mark.parametrize("from_remote", [True, False])
    def test_message_if_not_enabled(self, enable, from_remote):
        tc = TestClient(default_server_user=True)
        tc.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("compiler")})
        tc.run("create . -s=compiler.cppstd=17")
        if from_remote:
            tc.run("upload * -r=default -c")
            tc.run("remove * -c")
        arg = "-cc core.graph:compatibility_mode=optimized" if enable else ""
        tc.run(f"install --requires=pkg/0.1 -s=compiler.cppstd=14 {arg}")
        if not enable and from_remote:
            assert "A new experimental approach for binary compatibility detection is available" in tc.out
        else:
            assert "A new experimental approach for binary compatibility detection is available" not in tc.out

    @pytest.mark.parametrize("from_remote", [True, False])
    @pytest.mark.parametrize("update", [True, False])
    def test_compatibility_different_settings_per_context(self, from_remote, update):
        tc = TestClient(default_server_user=True)
        tc.save({"protobuf/conanfile.py": GenConanfile("protobuf", "1.0")
                .with_settings("compiler"),
                 "conanfile.py": GenConanfile("consumer", "1.0")
                .with_require("protobuf/1.0")
                .with_tool_requires("protobuf/1.0")
                 })
        tc.run("create protobuf -s=compiler.cppstd=14")
        if from_remote:
            tc.run("upload * -r=default -c")
            tc.run("remove * -c")
        update_arg = "--update" if update else ""
        tc.run(
            f"install . -s=compiler.cppstd=14 -s:b=compiler.cppstd=17 --build=missing {update_arg} "
            "-cc core.graph:compatibility_mode=optimized")

    @pytest.mark.parametrize("update", [True, False])
    def test_compatibility_different_settings_per_context_prevs(self, update):
        tc = TestClient(default_server_user=True)
        proto = GenConanfile("protobuf", "1.0").with_settings("compiler")
        proto.with_package_file("file.txt", env_var="MY_VAR")
        consumer = GenConanfile().with_requires("protobuf/1.0").with_tool_requires("protobuf/1.0")
        tc.save({"protobuf/conanfile.py": proto,
                 "conanfile.py": consumer})

        settings = "-s:a compiler=gcc -s:a compiler.version=9 -s:a compiler.libcxx=libstdc++"
        with environment_update({"MY_VAR": "value"}):
            tc.run(f"create protobuf {settings} -s=compiler.cppstd=14")
        tc.run("upload * -r=default -c")

        tc2 = TestClient(servers=tc.servers)
        tc2.save({"conanfile.py": consumer})
        update_arg = "--update" if update else ""
        tc2.run(f"install . {settings} -s=compiler.cppstd=14 -s:b=compiler.cppstd=17 {update_arg} "
                "-cc core.graph:compatibility_mode=optimized")
        tc2.assert_listed_binary({"protobuf/1.0": ("36d978cbb4dc35906d0fd438732d5e17cd1e388d",
                                                   "Download (default)")})

        with environment_update({"MY_VAR": "value2"}):
            tc.run(f"create protobuf {settings} -s=compiler.cppstd=14")
        tc.run("upload * -r=default -c")
        tc2.run(f"install . {settings} -s=compiler.cppstd=14 -s:b=compiler.cppstd=17 {update_arg} "
                "-cc core.graph:compatibility_mode=optimized")
        origin = "Cache" if not update else "Update (default)"
        tc2.assert_listed_binary({"protobuf/1.0": ("36d978cbb4dc35906d0fd438732d5e17cd1e388d",
                                                   origin)})

    def test_multi_remote(self):
        # https://github.com/conan-io/conan/issues/19342
        c = TestClient(servers={"r1": TestServer(), "r2": TestServer()},
                       inputs=["admin", "password"] * 2)

        c.save({"conanfile.py": GenConanfile("pkg", "0.1").with_settings("compiler")})
        c.run(f"create . -s=compiler.cppstd=14")

        c.run(f"upload * -r=r1 -c")
        c.run(f"upload * -r=r2 -c")
        c.run("remove * -c")
        c.run(f"install --requires=pkg/0.1 -s=compiler.cppstd=17 "
              "-cc core.graph:compatibility_mode=optimized")
        # It doesn't crash
        assert "pkg/0.1: Found 1 compatible configurations in remotes" in c.out


def test_compatibility_remove_cppstd():
    """ This test tries to reflect the following scenario:
    - User recently added compiler.cppstd to their settings
    - But up until now, no package was built with that setting
    - At the user's own risk, we can tell Conan to accept packages built without that setting
    """
    tc = TestClient()
    profile = textwrap.dedent("""
        [settings]
        compiler=gcc
        compiler.version=11
        compiler.libcxx=libstdc++11
        """)
    tc.save({"conanfile.py": GenConanfile("dep", "1.0").with_settings("compiler"),
             "profile": profile})
    tc.run("create . -pr=profile")
    dep_package_id = tc.created_package_id("dep/1.0")
    tc.run("install --requires=dep/1.0 -pr=profile -s=compiler.cppstd=17", assert_error=True)
    # We can't compile, because the dep is not compatible, it's looking for a package with cppstd
    assert "Missing prebuilt package for 'dep/1.0'" in tc.out

    # Let's create a compatibility extensions
    no_cppstd_compat = textwrap.dedent("""
        from conan.tools.scm import Version

        def no_cppstd_compat(conanfile):
            # Do we have the setting?
            cppstd_version = conanfile.settings.get_safe("compiler.cppstd")
            if cppstd_version is None:
                return []
            return [{"compiler.cppstd": None}]
        """)
    compat = tc.load_home("extensions/plugins/compatibility/compatibility.py")
    compat = "from no_cppstd_compat import no_cppstd_compat\n" + compat
    compat = compat.replace("# Append more factors for your custom compatibility rules here",
                            "factors.append(no_cppstd_compat(conanfile))")
    tc.save_home({"extensions/plugins/compatibility/no_cppstd_compat.py": no_cppstd_compat,
                  "extensions/plugins/compatibility/compatibility.py": compat})

    # Now we try again, this time app will find the compatible dep without cppstd
    tc.run("install --requires=dep/1.0 -pr=profile -s=compiler.cppstd=17")
    assert f"dep/1.0: Found compatible package '{dep_package_id}'" in tc.out



import os
from unittest.mock import patch, MagicMock

import pytest

from conan.errors import ConanException
from conan.internal.model.conf import Conf
from conan.test.utils.mocks import ConanFileMock, MockSettings
from conan.test.utils.test_files import temp_folder
from conan.tools.build import cmd_args_to_string
from conan.tools.files import save
from conan.tools.gnu import GnuToolchain


@pytest.fixture()
def cross_building_conanfile():
    settings_build = MockSettings({"os": "Linux",
                                   "arch": "x86_64",
                                   "compiler": "gcc",
                                   "compiler.version": "11",
                                   "compiler.libcxx": "libstdc++",
                                   "build_type": "Release"})
    settings_target = MockSettings({"os": "Android", "arch": "armv8"})
    settings = MockSettings({"os": "Emscripten", "arch": "wasm"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.settings_build = settings_build
    conanfile.settings_target = settings_target
    return conanfile


def test_get_gnu_triplet_for_cross_building():
    """
    Testing AutotoolsToolchainX and _get_gnu_triplet() function in case of
    having os=Windows and cross compiling
    """
    # Issue: https://github.com/conan-io/conan/issues/10139
    settings = MockSettings({"build_type": "Release",
                             "compiler": "gcc",
                             "compiler.version": "10.2",
                             "os": "Windows",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.settings_build = MockSettings({"os": "Solaris", "arch": "x86"})
    at = GnuToolchain(conanfile)
    assert at.configure_args["--host"] == "x86_64-w64-mingw32"
    assert at.configure_args["--build"] == "i686-solaris"


def test_get_toolchain_cppstd():
    settings = MockSettings({"build_type": "Release",
                             "compiler": "gcc",
                             "compiler.version": "9",
                             "compiler.cppstd": "20",
                             "os": "Linux",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.settings_build = settings
    at = GnuToolchain(conanfile)
    assert at.cppstd == "-std=c++2a"
    settings.values["compiler.version"] = "10"
    at = GnuToolchain(conanfile)
    assert at.cppstd == "-std=c++20"


@pytest.mark.parametrize("runtime, runtime_type, expected",
                         [("static", "Debug", "MTd"),
                          ("static", "Release", "MT"),
                          ("dynamic", "Debug", "MDd"),
                          ("dynamic", "Release", "MD")])
def test_msvc_runtime(runtime, runtime_type, expected):
    """
    Testing AutotoolsToolchainX with the msvc compiler adjust the runtime
    """
    # Issue: https://github.com/conan-io/conan/issues/10139
    settings = MockSettings({"build_type": "Release",
                             "compiler": "msvc",
                             "compiler.runtime": runtime,
                             "compiler.runtime_type": runtime_type,
                             "os": "Windows",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.settings_build = settings
    at = GnuToolchain(conanfile)
    expected_flag = "-{}".format(expected)
    assert at.msvc_runtime_flag == expected_flag
    env = at._environment.vars(conanfile)
    assert expected_flag in env["CFLAGS"]
    assert expected_flag in env["CXXFLAGS"]


@pytest.mark.parametrize("runtime", ["MTd", "MT", "MDd", "MD"])
def test_visual_runtime(runtime):
    """
    Testing AutotoolsToolchainX with the msvc compiler adjust the runtime
    """
    # Issue: https://github.com/conan-io/conan/issues/10139
    settings = MockSettings({"build_type": "Release" if "d" not in runtime else "Debug",
                             "compiler": "msvc",
                             "compiler.runtime": "static" if "MT" in runtime else "dynamic",
                             "compiler.runtime_type": "Release" if "d" not in runtime else "Debug",
                             "os": "Windows",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.settings_build = settings
    at = GnuToolchain(conanfile)
    expected_flag = "-{}".format(runtime)
    assert at.msvc_runtime_flag == expected_flag
    env = at._environment.vars(conanfile)
    assert expected_flag in env["CFLAGS"]
    assert expected_flag in env["CXXFLAGS"]


def test_get_gnu_triplet_for_cross_building_raise_error():
    """
    Testing AutotoolsToolchainX and _get_gnu_triplet() function raises an error in case of
    having os=Windows, cross compiling and not defined any compiler
    """
    # Issue: https://github.com/conan-io/conan/issues/10139
    settings = MockSettings({"build_type": "Release",
                             "os": "Windows",
                             "arch": "x86_64"})
    conanfile = ConanFileMock()
    conanfile.settings = settings
    conanfile.settings_build = MockSettings({"os": "Solaris", "arch": "x86"})
    with pytest.raises(ConanException) as conan_error:
        GnuToolchain(conanfile)
    msg = "'compiler' parameter for 'get_gnu_triplet()' is not specified and needed for os=Windows"
    assert msg == str(conan_error.value)


def test_compilers_mapping():
    autotools_mapping = {"c": "CC", "cpp": "CXX", "cuda": "NVCC", "fortran": "FC"}
    compilers = {"c": "path_to_c", "cpp": "path_to_cpp", "cuda": "path_to_cuda",
                 "fortran": "path_to_fortran"}
    settings = MockSettings({"build_type": "Release",
                             "os": "Windows",
                             "arch": "x86_64",
                             "compiler": "gcc"})
    conanfile = ConanFileMock()
    conanfile.conf = Conf()
    conanfile.conf.define("tools.build:compiler_executables", compilers)
    conanfile.settings = settings
    at = GnuToolchain(conanfile)
    env = at._environment.vars(conanfile)
    for compiler, env_var in autotools_mapping.items():
        assert env[env_var] == f"path_to_{compiler}"


def test_linker_scripts():
    conanfile = ConanFileMock()
    conanfile.conf = Conf()
    conanfile.conf.define("tools.build:linker_scripts", ["path_to_first_linker_script", "path_to_second_linker_script"])
    settings = MockSettings({"build_type": "Release",
                             "os": "Windows",
                             "compiler": "gcc",
                             "arch": "x86_64"})
    conanfile.settings = settings
    at = GnuToolchain(conanfile)
    env = at._environment.vars(conanfile)
    assert "-T'path_to_first_linker_script'" in env["LDFLAGS"]
    assert "-T'path_to_second_linker_script'" in env["LDFLAGS"]


def test_sysrootflag_qnx():
    """Even when no cross building it is adjusted because it could target a Mac version"""
    conanfile = ConanFileMock()
    conanfile.conf.define("tools.build:sysroot", "/path/to/sysroot")
    conanfile.settings = MockSettings(
        {"compiler": "qcc",
         "build_type": "Debug",
         "os": "Linux",
         "arch": "x86_64"})
    conanfile.settings_build = conanfile.settings
    be = GnuToolchain(conanfile)
    expected = "-Wc,-isysroot,/path/to/sysroot"
    assert be.sysroot_flag == expected


def test_update_or_prune_any_args(cross_building_conanfile):
    # Issue: https://github.com/conan-io/conan/issues/12642
    at = GnuToolchain(cross_building_conanfile)
    at.configure_args.update({
        "--with-cross-build": "my_path",
        "--something-host": "my_host",
        "--prefix": "/my/other/prefix"
    })
    new_configure_args = cmd_args_to_string(GnuToolchain._dict_to_list(at.configure_args))
    assert "--build=x86_64-linux-gnu" in new_configure_args
    assert "--host=wasm32-local-emscripten" in new_configure_args
    assert "--with-cross-build=my_path" in new_configure_args
    assert "--something-host=my_host" in new_configure_args
    assert "--prefix=/my/other/prefix" in new_configure_args
    # https://github.com/conan-io/conan/issues/12431
    at.configure_args.pop("--build")
    at.configure_args.pop("--host")
    new_configure_args = cmd_args_to_string(GnuToolchain._dict_to_list(at.configure_args))
    assert "--build=x86_64-linux-gnu" not in new_configure_args  # removed
    assert "--host=wasm32-local-emscripten" not in new_configure_args  # removed
    assert "--with-cross-build=my_path" in new_configure_args
    assert "--something-host=my_host" in new_configure_args
    # Update autoreconf_args
    at.autoreconf_args.pop("--force")
    new_autoreconf_args = cmd_args_to_string(GnuToolchain._dict_to_list(at.autoreconf_args))
    assert "'--force" not in new_autoreconf_args
    # Add new value to make_args
    at.make_args.update({"--new-complex-flag": "new-value"})
    at.make_args.update({"--new-empty-flag": ""})
    at.make_args.update({"--new-no-value-flag": None})
    new_make_args = cmd_args_to_string(GnuToolchain._dict_to_list(at.make_args))
    assert "--new-complex-flag=new-value" in new_make_args
    assert "--new-empty-flag=" in new_make_args
    assert "--new-no-value-flag" in new_make_args and "--new-no-value-flag=" not in new_make_args


@patch("conan.tools.gnu.gnutoolchain.VirtualBuildEnv")
def test_crossbuild_to_android(build_env_mock):
    """
    Issue related: https://github.com/conan-io/conan/issues/17441
    """
    buildvars = MagicMock()
    # VirtualBuildEnv defines these variables
    buildvars.vars.return_value = {"CC": "my-clang", "CXX": "my-clang++"}
    build_env_mock.return_value = buildvars

    conanfile = ConanFileMock()
    conanfile.settings = MockSettings({"os": "Android", "arch": "armv8", "os.api_level": "26r"})
    conanfile.settings_build = MockSettings({"os": "Macos", "arch": "armv8"})
    gnutc = GnuToolchain(conanfile)
    env_vars = gnutc.extra_env.vars(conanfile)
    assert env_vars.get("CC") is None
    assert env_vars.get("CXX") is None
    assert gnutc.triplets_info["host"]["triplet"] == "aarch64-linux-android"
    assert env_vars.get("LD") is None
    assert env_vars.get("STRIP") is None

    # Defining the ndk_path too
    ndk_path = temp_folder()
    ndk_bin = os.path.join(ndk_path, "toolchains", "llvm", "prebuilt", "darwin-x86_64", "bin")
    save(conanfile, os.path.join(ndk_bin, "ld"), "")
    conanfile.conf.define("tools.android:ndk_path", ndk_path)
    gnutc = GnuToolchain(conanfile)
    env_vars = gnutc.extra_env.vars(conanfile)
    assert env_vars.get("CC") is None
    assert env_vars.get("CXX") is None
    assert gnutc.triplets_info["host"]["triplet"] == "aarch64-linux-android"
    assert env_vars["LD"] == os.path.join(ndk_bin, "ld")  # exists
    assert env_vars["STRIP"] == os.path.join(ndk_bin, "llvm-strip")  # does not exist but appears


def test_gnu_toolchain_conf_extra_configure_args():
    """ Validate that tools.gnu:extra_configure_args are passed to the configure_args when
        building with GnuToolchain.
        The configure args should be passed as a list-like object.
    """
    conanfile = ConanFileMock()
    conanfile.settings = MockSettings({"os": "Linux", "arch": "x86_64"})
    conanfile.conf = Conf()
    conanfile.conf.define("tools.gnu:extra_configure_args", ["--foo", "--bar"])

    tc = GnuToolchain(conanfile)
    assert tc.configure_args["--foo"] is None
    assert tc.configure_args["--bar"] is None



import os

import fasteners

from conan.internal.errors import NotFoundException
from conan.internal.util.files import rmdir
from conans.server.utils.files import path_exists, relative_dirs


class ServerDiskAdapter:
    """Manage access to disk files with common methods required
    for conan operations"""
    def __init__(self, base_url, base_storage_path):
        """
        :param base_url Base url for generate urls to download and upload operations"""

        self.base_url = base_url
        # URLs are generated removing this base path
        self._store_folder = base_storage_path

    def get_file_list(self, absolute_path=""):
        if not path_exists(absolute_path, self._store_folder):
            raise NotFoundException("")
        paths = relative_dirs(absolute_path)
        abs_paths = [os.path.join(absolute_path, relpath) for relpath in paths]
        return abs_paths

    def delete_folder(self, path):
        """Delete folder from disk. Path already contains base dir"""
        if not path_exists(path, self._store_folder):
            raise NotFoundException("")
        rmdir(path)

    def path_exists(self, path):
        return os.path.exists(path)

    def read_file(self, path, lock_file):
        with fasteners.InterProcessLock(lock_file):
            with open(path) as f:
                return f.read()

    def write_file(self, path, contents, lock_file):
        with fasteners.InterProcessLock(lock_file):
            with open(path, "w") as f:
                f.write(contents)



from collections import OrderedDict

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import NO_SETTINGS_PACKAGE_ID, TestClient, TestServer


class TestRemoteChecks:

    def test_binary_defines_remote(self):
        servers = OrderedDict([("server1", TestServer()),
                               ("server2", TestServer()),
                               ("server3", TestServer())])
        client = TestClient(servers=servers, inputs=3*["admin", "password"])
        conanfile = """from conan import ConanFile
class Pkg(ConanFile):
    pass"""
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("upload pkg* -r=server1 --confirm")
        client.run("upload pkg* -r=server2 --confirm")

        # It takes the default remote
        client.run("remove * -c")

        # Exported recipe gets binary from default remote
        client.run("export . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("install --requires=pkg/0.1@lasote/testing")
        client.assert_listed_binary(
            {"pkg/0.1@lasote/testing": (NO_SETTINGS_PACKAGE_ID, "Download (server1)")})
        assert "pkg/0.1@lasote/testing: Retrieving package " \
                      "%s from remote 'server1'" % NO_SETTINGS_PACKAGE_ID in client.out

        # Explicit remote also defines the remote
        client.run("remove * -c")
        client.run("export . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("install --requires=pkg/0.1@lasote/testing -r=server2")
        client.assert_listed_binary(
            {"pkg/0.1@lasote/testing": (NO_SETTINGS_PACKAGE_ID, "Download (server2)")})
        assert "pkg/0.1@lasote/testing: Retrieving package " \
                      "%s from remote 'server2'" % NO_SETTINGS_PACKAGE_ID in client.out

        # Ordered search of binary works
        client.run("remove * -c")
        client.run("remove * -c -r=server1")
        client.run("export . --name=pkg --version=0.1 --user=lasote --channel=testing")
        client.run("install --requires=pkg/0.1@lasote/testing")
        client.assert_listed_binary(
            {"pkg/0.1@lasote/testing": (NO_SETTINGS_PACKAGE_ID, "Download (server2)")})
        assert "pkg/0.1@lasote/testing: Retrieving package " \
                      "%s from remote 'server2'" % NO_SETTINGS_PACKAGE_ID in client.out

        # Download recipe and binary from the remote2 by iterating
        client.run("remove * -c")
        client.run("remove * -c -r=server1")
        client.run("install --requires=pkg/0.1@lasote/testing")
        client.assert_listed_binary(
            {"pkg/0.1@lasote/testing": (NO_SETTINGS_PACKAGE_ID, "Download (server2)")})
        assert "pkg/0.1@lasote/testing: Retrieving package " \
                      "%s from remote 'server2'" % NO_SETTINGS_PACKAGE_ID in client.out

    def test_binaries_from_different_remotes(self):
        servers = OrderedDict()
        servers["server1"] = TestServer()
        servers["server2"] = TestServer()
        client = TestClient(servers=servers, inputs=2*["admin", "password"])
        conanfile = """from conan import ConanFile
class Pkg(ConanFile):
    options = {"opt": [1, 2, 3]}
"""
        client.save({"conanfile.py": conanfile})
        client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing -o pkg/*:opt=1")
        client.run("upload pkg* -r=server1 --confirm")
        client.run("remove *:* -c")
        client.run("create . --name=pkg --version=0.1 --user=lasote --channel=testing -o pkg/*:opt=2")
        package_id2 = client.created_package_id("pkg/0.1@lasote/testing")
        client.run("upload pkg* -r=server2 --confirm")
        client.run("remove *:* -c")

        # recipe is cached, takes binary from server2
        client.run("install --requires=pkg/0.1@lasote/testing -o pkg/*:opt=2 -r=server2")
        client.assert_listed_binary({"pkg/0.1@lasote/testing": (package_id2, "Download (server2)")})
        assert f"pkg/0.1@lasote/testing: Retrieving package {package_id2} " \
                      "from remote 'server2'" in client.out

        # Nothing to update
        client.run("install --requires=pkg/0.1@lasote/testing -o pkg/*:opt=2 -r=server2 -u")
        client.assert_listed_binary({"pkg/0.1@lasote/testing": (package_id2, "Cache")})

        # Build missing
        client.run("install --requires=pkg/0.1@lasote/testing -o pkg/*:opt=3 -r=server2", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'pkg/0.1@lasote/testing'" in client.out

        client.run("install --requires=pkg/0.1@lasote/testing -o pkg/*:opt=3", assert_error=True)
        assert "ERROR: Missing prebuilt package for 'pkg/0.1@lasote/testing'" in client.out


def test_version_range_multi_remote():
    """ captures the behavior of version-range resolution with multiple remotes:
    - normally the first occurrence stops: if a valid is found in the cache, it returns, if a
      valid is found in the first server, it is return
    - Using --update forces to really update to the latest version available, anywhere
    """
    servers = OrderedDict([("server1", TestServer()),
                           ("server2", TestServer()),
                           ("server3", TestServer())])
    client = TestClient(servers=servers, inputs=3*["admin", "password"])
    client.save({"conanfile.py": GenConanfile()})
    for i in (1, 2, 3):
        client.run(f"create . --name=pkg1 --version=1.{i}")
        client.run(f"upload pkg1/1.{i} -r=server{i}")
        client.run(f"create . --name=pkg2 --version=1.{i}")
        client.run(f"upload pkg2/1.{i} -r=server{4-i}")

    client.run("remove * -c")
    client.run("install  --requires=pkg1/[*] --requires=pkg2/[*]")
    # First found
    assert "pkg1/1.1#4d670581ccb765839f2239cc8dff8fbd - Downloaded (server1)" in client.out
    assert "pkg2/1.3#4d670581ccb765839f2239cc8dff8fbd - Downloaded (server1)" in client.out
    assert "pkg1/[*]: pkg1/1.1" in client.out
    assert "pkg2/[*]: pkg2/1.3" in client.out

    client.run("remove * -c")
    client.run("install  --requires=pkg1/[*] --requires=pkg2/[*] --update")
    # with --update, it guarantees the greatest version is found among remotes
    assert "pkg1/1.3#4d670581ccb765839f2239cc8dff8fbd - Downloaded (server3)" in client.out
    assert "pkg2/1.3#4d670581ccb765839f2239cc8dff8fbd - Downloaded (server1)" in client.out
    assert "pkg1/[*]: pkg1/1.3" in client.out
    assert "pkg2/[*]: pkg2/1.3" in client.out



import json
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_divergent_cppstd_build_host():

    c = TestClient()
    
    conanfile = textwrap.dedent("""
        [requires]
        waterfall/1.0
        [tool_requires]
        rainbow/1.0
        """)

    c.save({"waterfall/conanfile.py": GenConanfile("waterfall", "1.0").with_settings("compiler"),
            "rainbow/conanfile.py": GenConanfile("rainbow", "1.0").with_settings("compiler")
                                                .with_requires("waterfall/1.0"),
            "conanfile.txt": conanfile})


    c.run("export waterfall")
    c.run("export rainbow")
    c.run(f"install . --build=missing -s compiler.cppstd=14 -s:b compiler.cppstd=17 --format=json", redirect_stdout="graph.json")
    graph = json.loads(c.load("graph.json"))


    # waterfall is twice in the graph: as a direct host dependency, and an indirect build dependency
    assert graph['graph']['nodes']['1']['ref'] == "waterfall/1.0#821e924dcef2f185dd651e6d434f9f95"
    assert graph['graph']['nodes']['1']['context'] == "host"

    assert graph['graph']['nodes']['3']['ref'] == "waterfall/1.0#821e924dcef2f185dd651e6d434f9f95"
    assert graph['graph']['nodes']['3']['context'] == "build"

    # is this the right behaviour?
    # without the compatibility plugin, we would require two different package_id in the same graph,
    # but because of compatibility plugin, Conan graph uses a "compatible" package_id in the build context
    # rather than the exact one
    assert graph['graph']['nodes']['1']['package_id'] == graph['graph']['nodes']['3']['package_id']




import json
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_basic_inspect():
    t = TestClient()
    t.save({"foo/conanfile.py": GenConanfile().with_name("foo").with_shared_option()})
    t.run("inspect foo/conanfile.py")
    lines = t.out.splitlines()
    assert lines == ['default_options:',
                     '    shared: False',
                     'generators: []',
                     'label: ',
                     'languages: []',
                     'name: foo',
                     'options:',
                     '    shared: False',
                     'options_definitions:',
                     "    shared: ['True', 'False']",
                     'package_type: None',
                     'requires: []',
                     'revision_mode: hash',
                     'vendor: False'
                     ]


def test_options_description():
    t = TestClient()
    conanfile = textwrap.dedent("""\
        from conan import ConanFile
        class Pkg(ConanFile):
            options = {"shared": [True, False, None], "fpic": [True, False, None]}
            options_description = {"shared": "Some long explanation about shared option",
                                   "fpic": "Yet another long explanation of fpic"}
            """)
    t.save({"foo/conanfile.py": conanfile})
    t.run("inspect foo/conanfile.py")
    assert "shared: Some long explanation about shared option" in t.out
    assert "fpic: Yet another long explanation of fpic" in t.out


def test_missing_conanfile():
    t = TestClient()
    t.run("inspect missing/conanfile.py", assert_error=True)
    assert "Conanfile not found at" in t.out


def test_dot_and_folder_conanfile():
    t = TestClient()
    t.save({"conanfile.py": GenConanfile().with_name("foo")})
    t.run("inspect")
    assert 'name: foo' in t.out
    t.save({"foo/conanfile.py": GenConanfile().with_name("foo")}, clean_first=True)
    t.run("inspect foo")
    assert 'name: foo' in t.out


def test_inspect_understands_setname():
    tc = TestClient()
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Pkg(ConanFile):
        settings = "os", "arch"
        def set_name(self):
            self.name = "foo"

        def set_version(self):
            self.version = "1.0"
    """)

    tc.save({"conanfile.py": conanfile})
    tc.run("inspect .")
    assert "foo" in tc.out
    assert "1.0" in tc.out


def test_normal_inspect():
    tc = TestClient()
    tc.run("new basic -d name=pkg -d version=1.0")
    tc.run("inspect .")
    assert tc.out.splitlines() == ['description: A basic recipe',
                                   'generators: []',
                                   'homepage: <Your project homepage goes here>',
                                   'label: ',
                                   'languages: []',
                                   'license: <Your project license goes here>',
                                   'name: pkg',
                                   'options:',
                                   'options_definitions:',
                                   'package_type: None',
                                   'requires: []',
                                   'revision_mode: hash',
                                   'vendor: False',
                                   'version: 1.0']


def test_empty_inspect():
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Pkg(ConanFile):
        pass""")
    tc = TestClient()
    tc.save({"conanfile.py": conanfile})
    tc.run("inspect . -f json")


def test_basic_new_inspect():
    tc = TestClient()
    tc.run("new basic")
    tc.run("inspect . -f json")

    tc.run("new cmake_lib -d name=pkg -d version=1.0 -f")
    tc.run("inspect . -f json")


def test_requiremens_inspect():
    tc = TestClient()
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Pkg(ConanFile):
        requires = "zlib/1.2.13"
        license = "MIT", "Apache"
    """)
    tc.save({"conanfile.py": conanfile})
    tc.run("inspect .")
    assert ['generators: []',
            'label: ',
            'languages: []',
            "license: ['MIT', 'Apache']",
            'options:',
            'options_definitions:',
            'package_type: None',
            "requires: [{'ref': 'zlib/1.2.13', 'require': 'zlib/1.2.13', 'run': False, "
            "'libs': True, 'skip': False, 'test': False, 'force': False, 'direct': True, 'build': "
            "False, 'transitive_headers': None, 'transitive_libs': None, 'headers': "
            "True, 'package_id_mode': None, 'visible': True}]",
            'revision_mode: hash',
            'vendor: False'] == tc.out.splitlines()


def test_pythonrequires_remote():
    tc = TestClient(default_server_user=True)
    pyrequires = textwrap.dedent("""
    from conan import ConanFile

    class MyBase:
        def set_name(self):
            self.name = "my_company_package"

    class PyReq(ConanFile):
        name = "pyreq"
        version = "1.0"
        package_type = "python-require"
    """)
    tc.save({"pyreq/conanfile.py": pyrequires})
    tc.run("create pyreq/")
    tc.run("upload pyreq/1.0 -r default")
    tc.run("search * -r default")
    assert "pyreq/1.0" in tc.out
    tc.run("remove * -c")
    conanfile = textwrap.dedent("""
    from conan import ConanFile

    class Pkg(ConanFile):
        python_requires = "pyreq/1.0"
        python_requires_extend = "pyreq.MyBase"

        def set_version(self):
            self.version = "1.0"
    """)
    tc.save({"conanfile.py": conanfile})
    # Not specifying the remote also works
    tc.run("inspect .")
    assert "pyreq/1.0: Downloaded recipe revision 0ca726ab0febe1100901fffb27dc421f" in tc.out
    assert "name: my_company_package" in tc.out
    assert "version: 1.0" in tc.out
    # It now finds it on the cache, because it was downloaded
    tc.run("inspect . -nr")
    assert "name: my_company_package" in tc.out
    assert "version: 1.0" in tc.out
    assert "'recipe': 'Cache'" in tc.out
    tc.run("remove pyreq/* -c")
    # And now no remotes fails
    tc.run("inspect . -nr", assert_error=True)
    assert "Cannot resolve python_requires 'pyreq/1.0': No remote defined" in tc.out


def test_serializable_inspect():
    tc = TestClient()
    tc.save({"conanfile.py": GenConanfile("a", "1.0")
            .with_requires("b/2.0")
            .with_setting("os")
            .with_option("shared", [True, False])
            .with_generator("CMakeDeps")})
    tc.run("inspect . --format=json")
    assert json.loads(tc.out)["name"] == "a"



import os
import textwrap

import pytest

from conan.tools.cmake.cmakedeps.cmakedeps import FIND_MODE_CONFIG, FIND_MODE_MODULE, \
    FIND_MODE_BOTH, FIND_MODE_NONE
from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


@pytest.mark.parametrize("cmake_find_mode", [FIND_MODE_CONFIG, FIND_MODE_MODULE,
                                             FIND_MODE_BOTH, FIND_MODE_NONE, None])
def test_reuse_with_modules_and_config(cmake_find_mode):
    t = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile

        class Conan(ConanFile):
            def package_info(self):
                {}
        """)

    if cmake_find_mode is not None:
        s = 'self.cpp_info.set_property("cmake_find_mode", "{}")'.format(cmake_find_mode)
        conanfile = conanfile.format(s)
    t.save({"conanfile.py": conanfile})
    t.run("create . --name=mydep --version=1.0")

    conanfile = GenConanfile().with_name("myapp").with_require("mydep/1.0")\
                                                 .with_generator("CMakeDeps")\
                                                 .with_settings("build_type", "os", "arch")
    t.save({"conanfile.py": conanfile})

    t.run("install . --output-folder=install -s os=Linux -s compiler=gcc -s compiler.version=6 "
          "-s compiler.libcxx=libstdc++11")

    def exists_config():
        return os.path.exists(os.path.join(t.current_folder, "install", "mydep-config.cmake"))

    def exists_module():
        return os.path.exists(os.path.join(t.current_folder, "install", "Findmydep.cmake"))

    if cmake_find_mode == FIND_MODE_CONFIG or cmake_find_mode is None:
        # None is default "config"
        assert exists_config()
        assert not exists_module()
    elif cmake_find_mode == FIND_MODE_MODULE:
        assert not exists_config()
        assert exists_module()
    elif cmake_find_mode == FIND_MODE_BOTH:
        assert exists_config()
        assert exists_module()
    elif cmake_find_mode == FIND_MODE_NONE:
        assert not exists_config()
        assert not exists_module()



import os
import textwrap

from conan.test.assets.cmake import gen_cmakelists
from conan.test.assets.sources import gen_function_cpp, gen_function_h
from conan.test.utils.file_server import TestFileServer
from conan.test.utils.test_files import temp_folder
from conan.test.utils.tools import TestClient, zipdir
from conan.internal.util.files import save_files, sha256sum


class TestLocalRecipeIndexNew:
    def test_conan_new_local_recipes_index(self):
        # Setup the release pkg0.1.zip http server
        file_server = TestFileServer()
        zippath = os.path.join(file_server.store, "pkg0.1.zip")
        repo_folder = temp_folder()
        cmake = gen_cmakelists(libname="pkg", libsources=["pkg.cpp"], install=True,
                               public_header="pkg.h")
        save_files(repo_folder, {"pkg/CMakeLists.txt": cmake,
                                 "pkg/pkg.h": gen_function_h(name="pkg"),
                                 "pkg/pkg.cpp": gen_function_cpp(name="pkg")})
        zipdir(repo_folder, zippath)
        sha256 = sha256sum(zippath)
        url = f"{file_server.fake_url}/pkg0.1.zip"

        c0 = TestClient()
        c0.servers["file_server"] = file_server
        c0.run(f"new local_recipes_index -d name=pkg -d version=0.1 -d url={url} -d sha256={sha256}")
        # A local create is possible, and it includes a test_package
        c0.run("create recipes/pkg/all --version=0.1")
        assert "pkg: Release!" in c0.out
        remote_folder = c0.current_folder

        c = TestClient()
        c.servers["file_server"] = file_server
        c.run(f"remote add local '{remote_folder}'")
        c.run("new cmake_exe -d name=app -d version=0.1 -d requires=pkg/0.1")
        c.run("create . --version=0.1 --build=missing")
        assert "pkg: Release!" in c.out


class TestInRepo:
    def test_in_repo(self):
        """testing that it is possible to put a "recipes" folder inside a source repo, and
        use it as local-recipes-index repository, exporting the source from itself
        """
        repo_folder = temp_folder()
        cmake = gen_cmakelists(libname="pkg", libsources=["pkg.cpp"], install=True,
                               public_header="pkg.h")
        config_yml = textwrap.dedent("""\
            versions:
              "0.1":
                folder: all
            """)
        conanfile = textwrap.dedent("""\
            import os
            from conan import ConanFile
            from conan.tools.cmake import CMake, cmake_layout
            from conan.tools.files import copy

            class PkgRecipe(ConanFile):
                name = "pkg"
                package_type = "library"

                # Binary configuration
                settings = "os", "compiler", "build_type", "arch"
                options = {"shared": [True, False], "fPIC": [True, False]}
                default_options = {"shared": False, "fPIC": True}

                generators = "CMakeToolchain"

                def export_sources(self):
                    src = os.path.dirname(os.path.dirname(os.path.dirname(self.recipe_folder)))
                    copy(self, "*", src=src, dst=self.export_sources_folder, excludes=["recipes*"])

                def config_options(self):
                    if self.settings.os == "Windows":
                        self.options.rm_safe("fPIC")

                def configure(self):
                    if self.options.shared:
                        self.options.rm_safe("fPIC")

                def layout(self):
                    cmake_layout(self)

                def build(self):
                    cmake = CMake(self)
                    cmake.configure()
                    cmake.build()

                def package(self):
                    cmake = CMake(self)
                    cmake.install()

                def package_info(self):
                    self.cpp_info.libs = [self.name]
            """)

        save_files(repo_folder, {"recipes/pkg/config.yml": config_yml,
                                 "recipes/pkg/all/conanfile.py": conanfile,
                                 "CMakeLists.txt": cmake,
                                 "pkg.h": gen_function_h(name="pkg"),
                                 "pkg.cpp": gen_function_cpp(name="pkg")})

        c = TestClient()
        c.run(f"remote add local '{repo_folder}'")
        c.run("new cmake_exe -d name=app -d version=0.1 -d requires=pkg/0.1")
        c.run("create . --build=missing")
        assert "pkg: Release!" in c.out

        # Of course the recipe can also be created locally
        path = os.path.join(repo_folder, "recipes/pkg/all")
        c.run(f'create "{path}" --version=0.1')
        assert "pkg/0.1: Created package" in c.out

        # Finally lets remove the remote, check that the clone is cleared
        c.run('remote remove local')
        assert "Removing temporary files for 'local' local-recipes-index remote" in c.out

    def test_not_found(self):
        """testing that the correct exception is raised when a recipe is not found
        """
        repo1_folder = temp_folder()
        repo2_folder = temp_folder()
        config_yml = textwrap.dedent("""\
            versions:
              "0.1":
                folder: all
            """)
        conanfile = textwrap.dedent("""\
            import os
            from conan import ConanFile
            from conan.tools.files import copy

            class PkgRecipe(ConanFile):
                name = "pkg"

                def export_sources(self):
                    copy(self, "*", src=self.recipe_folder, dst=self.export_sources_folder)
            """)

        save_files(repo2_folder, {"recipes/pkg/config.yml": config_yml,
                                  "recipes/pkg/all/conanfile.py": conanfile,
                                  "recipes/pkg/all/pkg.h": gen_function_h(name="pkg")})

        c = TestClient()
        c.run(f"remote add local1 '{repo1_folder}'")
        c.run(f"remote add local2 '{repo2_folder}'")
        c.run("install --requires=pkg/0.1 --build=missing")
        assert "Install finished successfully" in c.out



import json
import os
import platform
import sys
from contextlib import contextmanager

import pytest
from unittest.mock import patch, MagicMock

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.env import environment_update
from conan.test.utils.tools import TestClient

_sbom_zlib_1_2_11 = """
{
  "components" : [ {
    "author" : "<Put your name here> <And your email here>",
    "bom-ref" : "pkg:conan/zlib@1.2.11?rref=6754320047c5dd54830baaaf9fc733c4",
    "description" : "<Description of zlib package here>",
    "licenses" : [ {
      "license" : {
        "name" : "<Put the package license here>"
      }
    } ],
    "name" : "zlib",
    "purl" : "pkg:conan/zlib@1.2.11",
    "type" : "library",
    "version" : "1.2.11"
  } ],
  "dependencies" : [ {
    "ref" : "pkg:conan/zlib@1.2.11?rref=6754320047c5dd54830baaaf9fc733c4"
  } ],
  "metadata" : {
    "component" : {
      "author" : "<Put your name here> <And your email here>",
      "bom-ref" : "pkg:conan/zlib@1.2.11?rref=6754320047c5dd54830baaaf9fc733c4",
      "name" : "zlib/1.2.11",
      "type" : "library"
    },
    "timestamp" : "2025-06-10T08:13:11Z",
    "tools" : [ {
      "externalReferences" : [ {
        "type" : "website",
        "url" : "https://github.com/conan-io/conan"
      } ],
      "name" : "Conan-io"
    } ]
  },
  "serialNumber" : "urn:uuid:4f7ce240-4c6d-4a87-bfb6-78d0fbc839e3",
  "bomFormat" : "CycloneDX",
  "specVersion" : "1.4",
  "version" : 1
}
"""


@contextmanager
def proxy_response(status, data, retry_after=60):
    with patch("conan.api.conan_api.ConanAPI._ApiHelpers.requester") as conanRequesterMock:
        return_status = MagicMock()
        return_status.status_code = status
        return_status.json = MagicMock(return_value=data)
        return_status.headers = {"retry-after": retry_after}
        conanRequesterMock.post = MagicMock(return_value=return_status)

        yield conanRequesterMock, return_status


def test_conan_audit_proxy():
    successful_response = {
        "data": {
            "query": {
                "vulnerabilities": {
                    "totalCount": 1,
                    "edges": [
                        {
                            "node": {
                                "name": "CVE-2023-45853",
                                "description": "Zip vulnerability",
                                "severity": "Critical",
                                "cvss": {
                                    "preferredBaseScore": 8.9
                                },
                                "aliases": [
                                    "CVE-2023-45853",
                                    "JFSA-2023-000272529"
                                ],
                                "advisories": [
                                    {
                                        "name": "CVE-2023-45853"
                                    },
                                    {
                                        "name": "JFSA-2023-000272529"
                                    }
                                ],
                                "references": [
                                    "https://pypi.org/project/pyminizip/#history",
                                ]
                            }
                        }
                    ]
                }
            }
        },
        "error": None
    }

    tc = TestClient(light=True, default_server_user=True)

    tc.save({"conanfile.py": GenConanfile("zlib", "1.2.11"),
             "sbom.cdx.json": _sbom_zlib_1_2_11})
    tc.run("create . --lockfile-out=conan.lock")
    tc.run("upload * -c -r=default")

    tc.run("list * -r=default -f=json", redirect_stdout="pkglist_remote.json")

    tc.run("list '*' -f=json", redirect_stdout="pkglist.json")

    tc.run("audit list zlib/1.2.11", assert_error=True)
    assert "Authentication required for the CVE provider: 'conancenter" in tc.out

    tc.run("audit provider auth conancenter --token=valid_token")

    with proxy_response(200, successful_response):
        tc.run("audit list zlib/1.2.11")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list -l=pkglist.json")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list -l=pkglist_remote.json -r=default")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list --lockfile=conan.lock")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list --sbom=sbom.cdx.json")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit scan --requires=zlib/1.2.11")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.save({"conanfile.txt": "[requires]\nzlib/1.2.11\n"}, clean_first=True)
        tc.run("audit scan")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit scan . --requires=zlib/1.2.11", assert_error=True)
        assert "--requires and --tool-requires arguments are incompatible with [path] '.' argument" in tc.out

    # Now some common errors, like rate limited or missing lib, but it should not fail!
    with proxy_response(429, {"error": "Rate limit exceeded"}):
        tc.run("audit list zlib/1.2.11", assert_error=True)
        assert "You have exceeded the number of allowed requests" in tc.out
        assert "The limit will reset in 1 minute" in tc.out

    with proxy_response(429, {"error": "Rate limit exceeded"}, retry_after=2 * 3600):
        tc.run("audit list zlib/1.2.11", assert_error=True)
        assert "You have exceeded the number of allowed requests" in tc.out
        assert "The limit will reset in 2 hours and 0 minute" in tc.out

    with proxy_response(400, {"error": "Not found"}):
        # Not finding a package should not be an error
        tc.run("audit list zlib/1.2.11")
        assert "Package 'zlib/1.2.11' not scanned: Not found." in tc.stdout

    with proxy_response(403, {"error": "Error not shown"}):
        tc.run("audit list zlib/1.2.11", assert_error=True)
        assert "ERROR: Authentication error (403)" in tc.out
        assert "Error not shown" not in tc.out

    with proxy_response(500, {"error": "Internal error"}):
        tc.run("audit list zlib/1.2.11", assert_error=True)
        assert "Internal server error (500)" in tc.out

    with proxy_response(405, {"error": "Method not allowed"}):
        tc.run("audit list zlib/1.2.11", assert_error=True)
        assert "Error in zlib/1.2.11 (405)" in tc.out

    tc.run("audit provider add myprivate --url=foo --type=private --token=valid_token")

    tc.run("audit provider list")
    assert "(type: conan-center-proxy)" in tc.out
    assert "(type: private)" in tc.out

    tc.run("audit provider remove conancenter")
    tc.run("audit list zlib/1.2.11", assert_error=True)
    assert ("ERROR: Provider 'conancenter' not found. Please specify a valid provider name or add "
            "it using: 'conan audit provider add conancenter --url=https://audit.conan.io/ "
            "--type=conan-center-proxy --token=<token>'") in tc.out
    assert "If you don't have a valid token, register at: https://audit.conan.io/register." in tc.out

    if platform.system() != "Windows":
        providers_stat = os.stat(os.path.join(tc.cache_folder, "audit_providers.json"))
        # Assert that only the current user can read/write the file
        assert providers_stat.st_uid == os.getuid()
        assert providers_stat.st_gid == os.getgid()
        assert providers_stat.st_mode & 0o777 == 0o600


def test_conan_audit_private():
    successful_response = {
        "data": {
            "query": {
                "vulnerabilities": {
                    "totalCount": 1,
                    "edges": [
                        {
                            "node": {
                                "name": "CVE-2023-45853",
                                "description": "Zip vulnerability",
                                "severity": "Critical",
                                "cvss": {
                                    "preferredBaseScore": 8.9
                                },
                                "aliases": [
                                    "CVE-2023-45853",
                                    "JFSA-2023-000272529"
                                ],
                                "advisories": [
                                    {
                                        "name": "CVE-2023-45853"
                                    },
                                    {
                                        "name": "JFSA-2023-000272529"
                                    }
                                ],
                                "references": [
                                    "https://pypi.org/project/pyminizip/#history",
                                ]
                            }
                        }
                    ]
                }
            }
        }
    }

    tc = TestClient(light=True)

    tc.save({"conanfile.py": GenConanfile("zlib", "1.2.11"),
             "sbom.cdx.json": _sbom_zlib_1_2_11})
    tc.run("create . --lockfile-out=conan.lock")

    tc.run("list '*' -f=json", redirect_stdout="pkglist.json")

    # TODO: If the CLI does not allow tokenless provider, should this case not be handled?
    tc.run("audit provider add myprivate --url=foo --type=private --token=f")
    # Now, remove the token as if the user didn't set it manually in the json
    providers = json.loads(tc.load_home("audit_providers.json"))
    providers["myprivate"].pop("token", None)
    tc.save_home({"audit_providers.json": json.dumps(providers)})

    tc.run("audit list zlib/1.2.11 -p=myprivate", assert_error=True)
    assert "Missing authentication token for 'myprivate' provider" in tc.out

    tc.run("audit provider auth myprivate --token=valid_token")

    with proxy_response(200, successful_response):
        tc.run("audit list zlib/1.2.11 -p=myprivate")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list -l=pkglist.json -p=myprivate")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list --lockfile=conan.lock -p=myprivate")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit list --sbom=sbom.cdx.json -p=myprivate")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

        tc.run("audit scan --requires=zlib/1.2.11  -p=myprivate")
        assert "zlib/1.2.11 1 vulnerability found" in tc.out

    # Now some common errors, like rate limited or missing lib, but it should not fail!
    with proxy_response(400, {"errors": [{"message": "Ref not found"}]}):
        # Not finding a package should not be an error
        tc.run("audit list zlib/1.2.11 -p=myprivate")
        assert "Package 'zlib/1.2.11' not scanned: Not found." in tc.stdout

    with proxy_response(403, {"errors": [{"message": "Authentication error"}]}):
        tc.run("audit list zlib/1.2.11 -p=myprivate")
        assert "Unknown error" in tc.out

    with proxy_response(500, {"errors": [{"message": "Internal error"}]}):
        tc.run("audit list zlib/1.2.11 -p=myprivate")
        assert "Unknown error" in tc.out

    with proxy_response(405, {"errors": [{"message": "Method not allowed"}]}):
        tc.run("audit list zlib/1.2.11 -p=myprivate")
        assert "Unknown error" in tc.out

    with proxy_response(404, {"errors": [{"message": "Not found"}]}):
        tc.run("audit list zlib/1.2.11 -p=myprivate")
        assert "An error occurred while connecting to the 'myprivate' provider" in tc.out



@pytest.mark.skipif(sys.version_info < (3, 10),
                    reason="Strict Base64 validation introduced in Python 3.10")
def test_conan_audit_corrupted_token():
    tc = TestClient(light=True)

    json_path = os.path.join(tc.cache_folder, "audit_providers.json")
    with open(json_path, "r") as f:
        data = json.load(f)

    # this is not a valid base64 string and will raise an exception
    data["conancenter"]["token"] = "corrupted_token"

    with open(json_path, "w") as f:
        json.dump(data, f)
    tc.run("audit list zlib/1.2.11", assert_error=True)
    assert "Invalid token format for provider 'conancenter'. The token might be corrupt." in tc.out


def test_audit_list_conflicting_args():
    tc = TestClient(light=True)
    tc.save({"pkglist.json": '{"Local Cache": {"zlib/1.2.11": {}}}'})
    tc.run("audit list zlib/1.2.11 -l=pkglist.json", assert_error=True)
    assert "argument -l/--list: not allowed with argument reference" in tc.out


def test_audit_provider_add_missing_url():
    tc = TestClient(light=True)
    tc.run("audit provider add myprivate --type=private --token=valid_token", assert_error=True)
    assert "Name, URL and type are required to add a provider" in tc.out


def test_audit_provider_remove_nonexistent():
    tc = TestClient(light=True)
    tc.run("audit provider remove nonexistingprovider", assert_error=True)
    assert "Provider 'nonexistingprovider' not found" in tc.out


def test_audit_list_missing_arguments():
    tc = TestClient(light=True)
    tc.run("audit list", assert_error=True)
    assert "one of the arguments reference" in tc.out


def test_audit_provider_env_credentials_with_proxy(monkeypatch):
    tc = TestClient(light=True)
    # Authenticate the provider with an old token to verify that the env variable overrides it
    tc.run("audit provider auth conancenter --token=old_token")

    captured_headers = {}

    def fake_post(url, headers, json):  # noqa
        # Capture the headers used in the request
        captured_headers.update(headers)
        response = MagicMock()
        response.status_code = 200
        response.json.return_value = {
            "data": {"query": {"vulnerabilities": {"totalCount": 0, "edges": []}}}
        }
        response.headers = {"retry-after": 60}
        return response

    with environment_update({"CONAN_AUDIT_PROVIDER_TOKEN_CONANCENTER": "env_token_value"}):
        with patch("conan.api.conan_api.ConanAPI._ApiHelpers.requester",
                   new_callable=MagicMock) as requester_mock:
            requester_mock.post = fake_post
            tc.run("audit list zlib/1.2.11")

    # Verify that the Authorization header uses the token from the environment variable
    assert captured_headers.get("Authorization") == "Bearer env_token_value"


def test_audit_global_error_exception():
    """
    Test that a global error returned by the provider results in ConanException
    raised by the formatter, using the 'details' field.
    """
    tc = TestClient(light=True)
    tc.run("audit provider auth conancenter --token=valid_token")

    mock_provider_result = {
        "data": {},
        "conan_error": "Fatal error."
    }

    with patch("conan.api.conan_api.AuditAPI.list", return_value=mock_provider_result):
        tc.run("audit list zlib/1.2.11", assert_error=True)
        assert "ERROR: Fatal error." in tc.out

        tc.run("audit list zlib/1.2.11 -f json", assert_error=True)
        assert "ERROR: Fatal error." in tc.out

        tc.run("audit list zlib/1.2.11 -f html", assert_error=True)
        assert "ERROR: Fatal error." in tc.out


@pytest.mark.parametrize("severity_level, threshold, should_fail", [
    (1.0, None, False),
    (8.9, None, False),
    (9.0, None, True),
    (9.1, None, True),
    (5.0, 5.1, False),
    (5.1, 5.1, True),
    (5.2, 5.1, True),
    (9.0, 11.0, False),
])
def test_audit_scan_threshold_error(severity_level, threshold, should_fail):
    """In case the severity level is equal or higher than the found for a CVE,
       the command should output the information as usual, and exit with non-success code error.
    """
    successful_response = {
        "data": {
            "query": {
                "vulnerabilities": {
                    "totalCount": 1,
                    "edges": [
                        {
                            "node": {
                                "name": "CVE-2023-45853",
                                "description": "Zip vulnerability",
                                "severity": "Critical",
                                "cvss": {
                                    "preferredBaseScore": severity_level
                                },
                                "aliases": [
                                    "CVE-2023-45853",
                                    "JFSA-2023-000272529"
                                ],
                                "advisories": [
                                    {
                                        "name": "CVE-2023-45853"
                                    },
                                    {
                                        "name": "JFSA-2023-000272529"
                                    }
                                ],
                                "references": [
                                    "https://pypi.org/project/pyminizip/#history",
                                ]
                            }
                        }
                    ]
                }
            }
        }
    }

    tc = TestClient(light=True)

    tc.save({"conanfile.py": GenConanfile("foobar", "0.1.0")})
    tc.run("export .")
    tc.run("audit provider auth conancenter --token=valid_token")

    with proxy_response(200, successful_response):
        severity_param = "" if threshold is None else f"-sl {threshold}"
        tc.run(f"audit scan --requires=foobar/0.1.0 {severity_param}", assert_error=should_fail)
        assert "foobar/0.1.0 1 vulnerability found" in tc.out
        assert f"CVSS: {severity_level}" in tc.out
        if should_fail:
            if threshold is None:
                threshold = "9.0"
            assert f"ERROR: The package foobar/0.1.0 has a CVSS score {severity_level} and exceeded the threshold severity level {threshold}" in tc.out

def test_parse_error_crash_when_no_edges():
    from conan.cli.commands.audit import _parse_error_threshold

    scan_result = {
        "data": {
            # this used to crash because dav1d not having vulnerabilities field
            "dav1d/1.4.3": {"error": {"details": "Package 'dav1d/1.4.3' not scanned: Not found."}},
            "zlib/1.2.11": {
                "vulnerabilities": {
                    "totalCount": 1,
                    "edges": [
                        {"node": {"cvss": {"preferredBaseScore": 7.0}}}
                    ]
                }
            }
        }
    }

    _parse_error_threshold(scan_result, error_level=5.0)
    assert "conan_error" in scan_result
    assert "zlib/1.2.11" in scan_result["conan_error"]
    assert "7.0" in scan_result["conan_error"]


@pytest.mark.parametrize("package_context", ["build", "host"])
@pytest.mark.parametrize("filter_context", ["build", "host", None])
def test_audit_scan_context_filter(package_context, filter_context):
    tc = TestClient(light=True)

    tc.save({"conanfile.py": GenConanfile("zlib", "1.2.11")})
    tc.run("export .")
    tc.run("audit provider auth conancenter --token=valid_token")

    requires = "requires" if package_context == "host" else "tool-requires"
    context = "" if filter_context is None else f"--context={filter_context}"

    with proxy_response(200, {}):
        tc.run(f"audit scan --{requires}=zlib/1.2.11 {context}")
        if filter_context is None or filter_context == package_context:
            assert "Requesting vulnerability info for: zlib/1.2.11" in tc.out
        else:
            assert "Requesting vulnerability info for: zlib/1.2.11" not in tc.out



import pytest

from conan.tools.apple import XcodeBuild
from conan.internal.model.conf import ConfDefinition
from conan.test.utils.mocks import ConanFileMock, MockSettings


@pytest.mark.parametrize("mode", ["quiet", None, "verbose"])
def test_verbosity_global(mode):
    conanfile = ConanFileMock()
    conf = ConfDefinition()
    if mode is not None:
        conf.loads(f"tools.build:verbosity={mode}")
    conanfile.conf = conf
    conanfile.settings = MockSettings({})
    xcodebuild = XcodeBuild(conanfile)

    xcodebuild.build("app.xcodeproj")
    if mode == "verbose":
        assert "-verbose" in conanfile.command
        assert "-quiet" not in conanfile.command
    elif mode == "quiet":
        assert "-verbose" not in conanfile.command
        assert "-quiet" in conanfile.command
    else:
        assert "-verbose" not in conanfile.command
        assert "-quiet" not in conanfile.command


def test_sdk_path():
    conanfile = ConanFileMock()
    conf = ConfDefinition()
    conf.loads("tools.apple:sdk_path=mypath")
    conanfile.conf = conf
    conanfile.settings = MockSettings({})
    xcodebuild = XcodeBuild(conanfile)
    xcodebuild.build("app.xcodeproj")
    assert "SDKROOT=mypath " in conanfile.command


def test_sdk():
    conanfile = ConanFileMock()
    conf = ConfDefinition()
    conf.loads("tools.apple:sdk_path=mypath")
    conanfile.conf = conf
    conanfile.settings = MockSettings({"os": "Macos",
                                       "os.sdk": "macosx"})
    xcodebuild = XcodeBuild(conanfile)
    xcodebuild.build("app.xcodeproj")
    # sdk_path takes preference
    assert "SDKROOT=mypath " in conanfile.command
    conf = ConfDefinition()
    conanfile.conf = conf
    xcodebuild = XcodeBuild(conanfile)
    xcodebuild.build("app.xcodeproj")
    assert "SDKROOT=macosx " in conanfile.command
    conanfile.settings = MockSettings({"os": "Macos",
                                       "os.sdk": "macosx",
                                       "os.sdk_version": "12.1"})
    xcodebuild = XcodeBuild(conanfile)
    xcodebuild.build("app.xcodeproj")
    assert "SDKROOT=macosx12.1 " in conanfile.command
    conanfile.settings = MockSettings({})
    xcodebuild = XcodeBuild(conanfile)
    xcodebuild.build("app.xcodeproj")
    assert "SDKROOT" not in conanfile.command


@pytest.mark.parametrize("os_name, os_version, expected_key", [
    ("Macos", "14.0", "MACOSX_DEPLOYMENT_TARGET"),
    ("iOS", "15.1", "IPHONEOS_DEPLOYMENT_TARGET"),
    ("watchOS", "8.0", "WATCHOS_DEPLOYMENT_TARGET"),
    ("tvOS", "15.0", "TVOS_DEPLOYMENT_TARGET"),
    ("visionOS", "1.0", "XROS_DEPLOYMENT_TARGET")
])
def test_deployment_target_and_quoting(os_name, os_version, expected_key):
    """
    Checks that the correct deployment target is passed and that paths are quoted.
    """
    conanfile = ConanFileMock()
    conanfile.settings = MockSettings({"os": os_name, "os.version": os_version})
    xcodebuild = XcodeBuild(conanfile)

    xcodebuild.build("My Project.xcodeproj", target="My Target")

    expected_arg = f" {expected_key}={os_version}"
    assert expected_arg in conanfile.command

    assert "-project 'My Project.xcodeproj'" in conanfile.command
    assert "-target 'My Target'" in conanfile.command


def test_no_deployment_target_if_version_is_missing():
    """
    Checks that the deployment target argument is not added if os.version is missing.
    """
    conanfile = ConanFileMock()
    conanfile.settings = MockSettings({"os": "Macos"})
    xcodebuild = XcodeBuild(conanfile)
    xcodebuild.build("app.xcodeproj")

    assert "MACOSX_DEPLOYMENT_TARGET" not in conanfile.command



import os
import re
import textwrap

from conan.test.assets.genconanfile import GenConanfile
from conan.test.utils.tools import TestClient


def test_editable_envvars():
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            name = "dep"
            version = "1.0"
            def layout(self):
                self.folders.source = "mysource"
                self.folders.build = "mybuild"
                self.layouts.source.runenv_info.append_path("MYRUNPATH", "mylocalsrc")
                self.layouts.build.buildenv_info.define_path("MYBUILDPATH", "mylocalbuild")

                self.layouts.package.buildenv_info.define_path("MYBUILDPATH", "mypkgbuild")
                self.layouts.package.runenv_info.append_path("MYRUNTPATH", "mypkgsrc")

            def package_info(self):
                self.buildenv_info.define("OTHERVAR", "randomvalue")
        """)

    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": GenConanfile().with_settings("os").with_requires("dep/1.0")})
    c.run("editable add dep  --name=dep --version=1.0")
    c.run("install pkg -s os=Linux -s:b os=Linux")
    build_path = os.path.join(c.current_folder, "dep", "mybuild", "mylocalbuild")
    buildenv = c.load("pkg/conanbuildenv.sh")
    assert f'export MYBUILDPATH="{build_path}"' in buildenv

    runenv = c.load("pkg/conanrunenv.sh")
    # The package_info() buildenv is aggregated, no prob
    # But don't use same vars
    assert 'export OTHERVAR="randomvalue"' in buildenv
    run_path = os.path.join(c.current_folder, "dep", "mysource", "mylocalsrc")
    assert f'export MYRUNPATH="$MYRUNPATH:{run_path}"' in runenv

    c.run("editable remove dep")
    c.run("create dep")
    c.run("install pkg -s os=Linux -s:b os=Linux")
    buildenv = c.load("pkg/conanbuildenv.sh")
    assert 'mypkgbuild' in buildenv
    assert "mylocalbuild" not in buildenv
    assert 'export OTHERVAR="randomvalue"' in buildenv
    runenv = c.load("pkg/conanrunenv.sh")
    assert 'mypkgsrc' in runenv
    assert "mylocalsrc" not in runenv


def test_editable_envvars_package_info():
    """ if the ``layout()`` defines cpp.build.runenv_info/conf_info and the ``package_info()``
    also defines them, the ``package_info()`` values have precedence. This is not symmetrical
    with ``cpp_info.includedirs/xxxdirs``. The recommended solution is to use
    ``self.layouts.package.runenvinfo/conf_info`` instead.
    """
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            name = "dep"
            version = "1.0"
            def layout(self):
                self.layouts.build.runenv_info.define("SOME_PATH", "mypath_layout")
                self.cpp.build.includedirs = ["mylayoutinclude"]
                self.layouts.build.conf_info.define("user:myconf", "mylayoutconf")

            def package_info(self):
                print("Running package_info!!")
                self.runenv_info.define("SOME_PATH", "mypath_pkginfo")
                self.cpp_info.includedirs = ["mypkginfoinclude"]
                self.conf_info.define("user:myconf", "mypkginfoconf")
        """)
    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            name = "pkg"
            version = "1.0"
            settings = "os", "build_type"
            requires = "dep/1.0"
            def generate(self):
                myconf = self.dependencies["dep"].conf_info.get("user:myconf")
                self.output.info(f"DEP CONFINFO {myconf}")
        """)

    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": pkg})
    c.run("editable add dep ")
    c.run("install pkg -s os=Linux -s:b os=Linux -g CMakeDeps")
    assert "conanfile.py (pkg/1.0): DEP CONFINFO mylayoutconf" in c.out
    cmake = c.load("pkg/dep-release-data.cmake")
    assert 'set(dep_INCLUDE_DIRS_RELEASE "${dep_PACKAGE_FOLDER_RELEASE}/mylayoutinclude")' in cmake
    runenv = c.load("pkg/conanrunenv-release.sh")
    assert f'export SOME_PATH="mypath_layout"' in runenv


def test_editable_envvars_package():
    """ This test shows that ``self.layouts.package.runenvinfo/conf_info`` works, ignored
    while in editable, but used when regular package
    """
    c = TestClient()
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            name = "dep"
            version = "1.0"
            def layout(self):
                self.layouts.build.runenv_info.define("SOME_VALUE", "mypath_layout")
                self.cpp.build.includedirs = ["mylayoutinclude"]
                self.layouts.build.conf_info.define("user:myconf", "mylayoutconf")

                self.layouts.package.runenv_info.define("SOME_VALUE", "mypath_pkginfo")
                self.cpp.package.includedirs = ["mypkginfoinclude"]
                self.layouts.package.conf_info.define("user:myconf", "mypkginfoconf")
        """)
    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            name = "pkg"
            version = "1.0"
            settings = "os", "build_type"
            requires = "dep/1.0"
            def generate(self):
                myconf = self.dependencies["dep"].conf_info.get("user:myconf")
                self.output.info(f"DEP CONFINFO {myconf}")
        """)

    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": pkg})
    c.run("editable add dep ")
    c.run("install pkg -s os=Linux -s:b os=Linux -g CMakeDeps")
    assert "conanfile.py (pkg/1.0): DEP CONFINFO mylayoutconf" in c.out
    cmake = c.load("pkg/dep-release-data.cmake")
    assert 'set(dep_INCLUDE_DIRS_RELEASE "${dep_PACKAGE_FOLDER_RELEASE}/mylayoutinclude")' in cmake
    runenv = c.load("pkg/conanrunenv-release.sh")
    assert f'export SOME_VALUE="mypath_layout"' in runenv

    c.run("editable remove dep")
    c.run("create dep")
    c.run("install pkg -s os=Linux -s:b os=Linux -g CMakeDeps")
    assert "conanfile.py (pkg/1.0): DEP CONFINFO mypkginfoconf" in c.out
    cmake = c.load("pkg/dep-release-data.cmake")
    assert 'set(dep_INCLUDE_DIRS_RELEASE "${dep_PACKAGE_FOLDER_RELEASE}/mypkginfoinclude")' in cmake
    runenv = c.load("pkg/conanrunenv-release.sh")
    assert f'export SOME_VALUE="mypath_pkginfo"' in runenv


def test_editable_conf():
    c = TestClient()
    # TODO: Define if we want conf.xxxx_path(), instead of (..., path=True) methods
    dep = textwrap.dedent("""
        from conan import ConanFile
        class Dep(ConanFile):
            def layout(self):
                self.folders.source = "mysource"
                self.folders.build = "mybuild"
                self.layouts.source.conf_info.append_path("user:myconf", "mylocalsrc")
                self.layouts.build.conf_info.append_path("user:myconf", "mylocalbuild")
                self.layouts.build.conf_info.update_path("user:mydictconf", {"a": "mypatha",
                                                                             "b": "mypathb"})
                self.layouts.build.conf_info.define_path("user:mydictconf2", {"c": "mypathc"})
        """)

    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            requires = "dep/1.0"
            def generate(self):
                conf = self.dependencies["dep"].conf_info.get("user:myconf")
                self.output.info(f"CONF: {conf}")
                dictconf = self.dependencies["dep"].conf_info.get("user:mydictconf", check_type=dict)
                self.output.info(f"CONFDICT: {dictconf}")
                dictconf2 = self.dependencies["dep"].conf_info.get("user:mydictconf2",
                                                                   check_type=dict)
                self.output.info(f"CONFDICT: {dictconf2}")
        """)
    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": pkg})
    c.run("editable add dep --name=dep --version=1.0")
    c.run("install pkg -s os=Linux -s:b os=Linux")
    out = str(c.out).replace("\\\\", "\\")
    conf_source = os.path.join(c.current_folder, "dep", "mybuild", "mylocalbuild")
    conf_build = os.path.join(c.current_folder, "dep", "mysource", "mylocalsrc")
    confdict1 = os.path.join(c.current_folder, "dep", "mybuild", "mypatha")
    confdict2 = os.path.join(c.current_folder, "dep", "mybuild", "mypathc")
    assert conf_source in out
    assert conf_build in out
    assert confdict1 in out
    assert confdict2 in out


def test_editable_conf_tool_require_builtin():
    c = TestClient()
    dep = textwrap.dedent("""
        import os
        from conan import ConanFile
        class Dep(ConanFile):
            name = "androidndk"
            version = "1.0"
            def layout(self):
                self.layouts.build.conf_info.define_path("tools.android:ndk_path", "mybuild")
                self.layouts.package.conf_info.define_path("tools.android:ndk_path", "mypkg")
        """)

    pkg = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            tool_requires = "androidndk/1.0"
            def generate(self):
                ndk = self.conf.get("tools.android:ndk_path")
                self.output.info(f"NDK: {ndk}!!!")
        """)
    c.save({"dep/conanfile.py": dep,
            "pkg/conanfile.py": pkg})
    c.run("editable add dep")
    c.run("install pkg -s os=Linux -s:b os=Linux")
    ndk_path = os.path.join(c.current_folder, "dep", "mybuild")
    assert f"conanfile.py: NDK: {ndk_path}!!!" in c.out

    c.run("editable remove dep")
    c.run("create dep")
    c.run("install pkg -s os=Linux -s:b os=Linux")
    assert re.search("conanfile.py: NDK: .*mypkg!!!", str(c.out))



import textwrap

from conan.test.utils.tools import TestClient


def test_settings_definitions():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "os"
            def generate(self):
                definition = self.settings.possible_values()
                assert "os" in definition
                assert "compiler" not in definition
                assert "arch" not in definition
                self.output.info(", ".join(definition["os"]))
                ios = definition["os"]["iOS"]["version"]
                self.output.info(f"iOS: {ios}")

                os_definition = self.settings.os.possible_values()
                ios = os_definition["iOS"]["version"]
                self.output.info(f"iOS2: {ios}")
            """)
    c.save({"conanfile.py": conanfile})
    # New settings are there
    c.run("install . -s os=Linux")
    assert "conanfile.py: Windows, WindowsStore, WindowsCE, Linux," in c.out
    assert "conanfile.py: iOS: ['7.0', '7.1', '8.0'," in c.out
    assert "conanfile.py: iOS2: ['7.0', '7.1', '8.0'," in c.out


def test_settings_definitions_compiler():
    c = TestClient()
    conanfile = textwrap.dedent("""
        from conan import ConanFile
        class Pkg(ConanFile):
            settings = "compiler"
            def generate(self):
                definition = self.settings.compiler.version.possible_values()
                self.output.info("HOST: " + ", ".join(definition))
                definition = self.settings_build.compiler.version.possible_values()
                self.output.info("BUILD: " + ", ".join(definition))
                cppstds = self.settings.compiler.cppstd.possible_values()
                self.output.info("CPPSTDS: " + str(cppstds))
            """)
    profile = textwrap.dedent("""\
        [settings]
        compiler=msvc
        compiler.version=192
        compiler.runtime=dynamic
        """)
    c.save({"conanfile.py": conanfile,
            "profile": profile})
    # New settings are there
    c.run("install . -pr=profile -s:b compiler=gcc")
    assert "conanfile.py: HOST: 170, 180, 190, 191, 192" in c.out
    assert "conanfile.py: BUILD: 4.1, 4.4, 4.5, 4.6, 4.7," in c.out
    assert "conanfile.py: CPPSTDS: [None, '14', '17', '20', '23']" in c.out



import platform
import textwrap

import pytest

from conan.test.utils.tools import TestClient


@pytest.mark.skipif(platform.system() != "Linux", reason="SCons functional tests"
                                                         "only for Linux")
@pytest.mark.tool("scons")
def test_sconsdeps():
    client = TestClient(path_with_spaces=False)

    conanfile = textwrap.dedent("""\
        import os

        from conan import ConanFile
        from conan.tools.files import copy


        class helloConan(ConanFile):
            name = "hello"
            version = "1.0"
            settings = "os", "compiler", "build_type", "arch"

            exports_sources = "src/*"

            # TODO: check what would be the correct layout and how to interact with
            # SCons scripts
            def layout(self):
                self.folders.source = "src"

            def build(self):
                debug_opt = '--debug-build' if self.settings.build_type == 'Debug' else ''
                self.run(f'scons -C {self.folders.source} {debug_opt}')

            def package(self):
                copy(self, pattern="*.h", dst=os.path.join(self.package_folder, "include"), src=os.path.join(self.source_folder),)
                copy(self, "*.lib", src=self.source_folder, dst=os.path.join(self.package_folder, "lib"), keep_path=False)
                copy(self, "*.a", src=self.source_folder, dst=os.path.join(self.package_folder, "lib"), keep_path=False)

            def package_info(self):
                self.cpp_info.libs = ["hello"]

        """)

    hello_cpp = textwrap.dedent("""\
        #include <iostream>
        #include "hello.h"

        void hello(){
            #ifdef NDEBUG
                std::cout << "Hello World Release!" <<std::endl;
            #else
                std::cout << "Hello World Debug!" <<std::endl;
            #endif
        }
        """)

    hello_h = textwrap.dedent("""\
        #pragma once

        #ifdef WIN32
          #define HELLO_EXPORT __declspec(dllexport)
        #else
          #define HELLO_EXPORT
        #endif

        HELLO_EXPORT void hello();
        """)

    sconscript = textwrap.dedent("""\
        import sys

        AddOption('--debug-build', action='store_true', help='debug build')

        env = Environment(TARGET_ARCH="x86_64")

        is_debug = GetOption('debug_build')
        is_release = not is_debug

        if not is_debug:
            env.Append(CPPDEFINES="NDEBUG")

        if is_debug:
            env.Append(CXXFLAGS = '-g -ggdb')
        else:
            env.Append(CXXFLAGS = '-O2')
            env.Append(LINKFLAGS = '-O2')

        env.Library("hello", "hello.cpp")
        """)

    sconstruct = textwrap.dedent("""\
        SConscript('SConscript', variant_dir='build', duplicate = False)
        """)

    t_sconscript = textwrap.dedent("""\
        import sys

        AddOption('--debug-build', action='store_true', help='debug build')

        env = Environment(TARGET_ARCH="x86_64")

        is_debug = GetOption('debug_build')
        is_release = not is_debug

        if not is_debug:
            env.Append(CPPDEFINES="NDEBUG")

        if is_debug:
            env.Append(CXXFLAGS = '-g -ggdb')
        else:
            env.Append(CXXFLAGS = '-O2')
            env.Append(LINKFLAGS = '-O2')

        build_path_relative_to_sconstruct = Dir('.').path

        conandeps = SConscript('./SConscript_conandeps')

        flags = conandeps["conandeps"]
        env.MergeFlags(flags)

        env.Program("main", "main.cpp")
        """)

    t_sconstruct = textwrap.dedent("""\
        SConscript('SConscript', variant_dir='build', duplicate = False)
        """)
    t_conanfile = textwrap.dedent("""\
        import os

        from conan import ConanFile

        class helloTestConan(ConanFile):
            settings = "os", "compiler", "build_type", "arch"
            generators = "SConsDeps"
            apply_env = False
            test_type = "explicit"

            def requirements(self):
                self.requires(self.tested_reference_str)

            def build(self):
                debug_opt = '--debug-build' if self.settings.build_type == 'Debug' else ''
                self.run(f'scons {debug_opt}')

            # TODO: check how to setup layout and scons
            def layout(self):
                self.folders.source = "."
                self.folders.generators = self.folders.source
                self.cpp.build.bindirs = ["build"]

            def test(self):
                cmd = os.path.join(self.cpp.build.bindirs[0], "main")
                self.run(cmd, env="conanrun")
        """)

    t_main_cpp = textwrap.dedent("""\
        #include "hello.h"

        int main() {
            hello();
        }
        """)

    client.save({"conanfile.py": conanfile,
                 "src/hello.cpp": hello_cpp,
                 "src/hello.h": hello_h,
                 "src/SConscript": sconscript,
                 "src/SConstruct": sconstruct,
                 "test_package/SConscript": t_sconscript,
                 "test_package/SConstruct": t_sconstruct,
                 "test_package/conanfile.py": t_conanfile,
                 "test_package/main.cpp": t_main_cpp,
                 })

    client.run("create .")
    assert "Hello World Release!" in client.out



from conan.api.model import PkgReference
from conan.api.model import RecipeReference


def get_package_ref(name, version, username, channel, package_id, revision, p_revision):
    ref = RecipeReference(name, version, username, channel, revision)
    return PkgReference(ref, package_id, p_revision)



from conan.test.utils.tools import TestClient


class TestSysroot:

    def test(self):
        client = TestClient()
        sysroot = """from conan import ConanFile
class Pkg(ConanFile):
    def package_info(self):
        self.cpp_info.sysroot = "HelloSysRoot"
"""
        client.save({"conanfile.py": sysroot})
        client.run("create . --name=sysroot --version=0.1 --user=user --channel=testing")

        conanfile = """from conan import ConanFile
class Pkg(ConanFile):
    requires = "sysroot/0.1@user/testing"
    def build(self):
        self.output.info("PKG SYSROOT: %s" % self.dependencies["sysroot"].cpp_info.sysroot)
    def package_info(self):
        self.cpp_info.sysroot = "HelloSysRoot"
"""
        test_conanfile = """from conan import ConanFile
class Pkg(ConanFile):
    def requirements(self):
        self.requires(self.tested_reference_str)
    def build(self):
        self.output.info("Test SYSROOT: %s"
                          % self.dependencies["sysroot"].cpp_info.sysroot)
    def test(self):
        pass
"""
        client.save({"conanfile.py": conanfile,
                     "test_package/conanfile.py": test_conanfile})
        client.run("create . --name=pkg --version=0.1 --user=user --channel=testing")
        assert "pkg/0.1@user/testing: PKG SYSROOT: HelloSysRoot" in client.out
        assert "pkg/0.1@user/testing (test package): Test SYSROOT: HelloSysRoot" in client.out

        client.run("install .")



import textwrap

import pytest

from conan.test.utils.tools import TestClient

simple = """
from conan import ConanFile
class Pkg(ConanFile):
    options = {"shared": [True, False],
               "header_only": [True, False]}
"""

pkg_type = """
from conan import ConanFile
class Pkg(ConanFile):
    package_type = "library"
    options = {"shared": [True, False],
               "header_only": [True, False]}
"""

remove = """
from conan import ConanFile
class Pkg(ConanFile):
    package_type = "library"
    options = {"shared": [True, False],
               "header_only": [True, False]}
    def configure(self):
        if self.options.header_only:
            self.options.rm_safe("shared")
"""


@pytest.mark.parametrize("conanfile", [simple, pkg_type, remove])
def test_auto_package_type(conanfile):
    c = TestClient(light=True)
    c.save({"conanfile.py": conanfile})
    c.run("graph info . --filter package_type")
    assert "package_type: static-library" in c.out
    c.run("graph info . --filter package_type -o shared=True")
    assert "The package_type will have precedence over the options" not in c.out
    assert "package_type: shared-library" in c.out
    c.run("graph info . --filter package_type -o shared=True -o header_only=False")
    assert "package_type: shared-library" in c.out
    c.run("graph info . --filter package_type -o header_only=True")
    assert "package_type: header-library" in c.out
    c.run("graph info . --filter package_type -o header_only=True -o shared=False")
    assert "package_type: header-library" in c.out

def test_package_type_and_header_library():
    """ Show that forcing a package_type and header_only=True does not change the package_type"""
    tc = TestClient(light=True)
    tc.save({"conanfile.py": textwrap.dedent("""
    from conan import ConanFile

    class Pkg(ConanFile):
        package_type = "static-library"
        options = {"header_only": [True, False]}

    """)})
    tc.run("graph info . --filter package_type -o &:header_only=False")
    assert "package_type: static-library" in tc.out
    assert "The package_type will have precedence over the options" in tc.out
    tc.run("graph info . --filter package_type -o &:header_only=True")
    assert "package_type: static-library" in tc.out
    assert "The package_type will have precedence over the options" in tc.out
